#!/usr/bin/env python
"""

    ruffus_template.py
                        [--log_file PATH]
                        [--verbose]
                        [--target_tasks]
                        [--jobs]
                        [--just_print]
                        [--flowchart]
                        [--key_legend_in_graph]
                        [--forced_tasks]

"""
import sys, os


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


if __name__ == '__main__':
    from optparse import OptionParser
    import StringIO

    parser = OptionParser(version="%prog 1.0", usage = "\n\n    %progs [options]")



    #
    #   general options: verbosity / logging
    #
    parser.add_option("-v", "--verbose", dest = "verbose",
                      action="count", default=0,
                      help="Print more verbose messages for each additional verbose level.")
    parser.add_option("-L", "--log_file", dest="log_file",
                      metavar="FILE",
                      type="string",
                      help="Name and path of log file")




    #
    #   pipeline
    #
    parser.add_option("-t", "--target_tasks", dest="target_tasks",
                        action="append",
                        default = list(),
                        metavar="JOBNAME",
                        type="string",
                        help="Target task(s) of pipeline.")
    parser.add_option("-j", "--jobs", dest="jobs",
                        default=1,
                        metavar="N",
                        type="int",
                        help="Allow N jobs (commands) to run simultaneously.")
    parser.add_option("-n", "--just_print", dest="just_print",
                        action="store_true", default=False,
                        help="Don't actually run any commands; just print the pipeline.")
    parser.add_option("--flowchart", dest="flowchart",
                        metavar="FILE",
                        type="string",
                        help="Don't actually run any commands; just print the pipeline "
                             "as a flowchart.")

    #
    #   Less common pipeline options
    #
    parser.add_option("--key_legend_in_graph", dest="key_legend_in_graph",
                        action="store_true", default=False,
                        help="Print out legend and key for dependency graph.")
    parser.add_option("--forced_tasks", dest="forced_tasks",
                        action="append",
                        default = list(),
                        metavar="JOBNAME",
                        type="string",
                        help="Pipeline task(s) which will be included even if they are up to date.")

    # get help string
    f =StringIO.StringIO()
    parser.print_help(f)
    helpstr = f.getvalue()
    (options, remaining_args) = parser.parse_args()


    #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
    #                                             #
    #   Change this if necessary                  #
    #                                             #
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    #
    #   Add names of mandatory options,
    #       strings corresponding to the "dest" parameter
    #       in the options defined above
    #
    mandatory_options = [ ]

    #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
    #                                             #
    #   Change this if necessary                  #
    #                                             #
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


    def check_mandatory_options (options, mandatory_options, helpstr):
        """
        Check if specified mandatory options have b een defined
        """
        missing_options = []
        for o in mandatory_options:
            if not getattr(options, o):
                missing_options.append("--" + o)

        if not len(missing_options):
            return

        raise Exception("Missing mandatory parameter%s: %s.\n\n%s\n\n" %
                        ("s" if len(missing_options) > 1 else "",
                         ", ".join(missing_options),
                         helpstr))
    check_mandatory_options (options, mandatory_options, helpstr)


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

from ruffus import *
from ruffus.ruffus_exceptions import JobSignalledBreak

#from json import dumps
#from collections import defaultdict



#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Functions


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Logger


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

if __name__ == '__main__':
    import logging
    import logging.handlers

    MESSAGE = 15
    logging.addLevelName(MESSAGE, "MESSAGE")

    def setup_std_logging (logger, log_file, verbose):
        """
        set up logging using programme options
        """
        class debug_filter(logging.Filter):
            """
            Ignore INFO messages
            """
            def filter(self, record):
                return logging.INFO != record.levelno

        class NullHandler(logging.Handler):
            """
            for when there is no logging
            """
            def emit(self, record):
                pass

        # We are interesting in all messages
        logger.setLevel(logging.DEBUG)
        has_handler = False

        # log to file if that is specified
        if log_file:
            handler = logging.FileHandler(log_file, delay=False)
            handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)6s - %(message)s"))
            handler.setLevel(MESSAGE)
            logger.addHandler(handler)
            has_handler = True

        # log to stderr if verbose
        if verbose:
            stderrhandler = logging.StreamHandler(sys.stderr)
            stderrhandler.setFormatter(logging.Formatter("    %(message)s"))
            stderrhandler.setLevel(logging.DEBUG)
            if log_file:
                stderrhandler.addFilter(debug_filter())
            logger.addHandler(stderrhandler)
            has_handler = True

        # no logging
        if not has_handler:
            logger.addHandler(NullHandler())


    #
    #   set up log
    #
    logger = logging.getLogger(module_name)
    setup_std_logging(logger, options.log_file, options.verbose)

    #
    #   Allow logging across Ruffus pipeline
    #
    def get_logger (logger_name, args):
        return logger

    from ruffus.proxy_logger import *
    (logger_proxy,
     logging_mutex) = make_shared_logger_and_proxy (get_logger,
                                                    module_name,
                                                    {})


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Pipeline


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#       Put pipeline code here







#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
if __name__ == '__main__':
    if options.just_print:
        pipeline_printout(sys.stdout, options.target_tasks, options.forced_tasks,
                            verbose=options.verbose)

    elif options.flowchart:
        pipeline_printout_graph (   open(options.flowchart, "w"),
                                    # use flowchart file name extension to decide flowchart format
                                    #   e.g. svg, jpg etc.
                                    os.path.splitext(options.flowchart)[1][1:],
                                    options.target_tasks,
                                    options.forced_tasks,
                                    no_key_legend   = not options.key_legend_in_graph)
    else:
        pipeline_run(options.target_tasks, options.forced_tasks,
                            multiprocess    = options.jobs,
                            logger          = stderr_logger,
                            verbose         = options.verbose)




#!/usr/bin/env python
import sys
import os
from setuptools import setup

if not sys.version_info[0:2] >= (2, 6):
    sys.stderr.write("Requires Python later than 2.6\n")
    sys.exit(1)

# quickly import the latest version of ruffus
sys.path.insert(0, os.path.abspath("."))
import ruffus.ruffus_version
sys.path.pop(0)

module_dependencies = []
#module_dependencies = ['multiprocessing>=2.6', 'simplejson']


setup(
        name='ruffus',
        version=ruffus.ruffus_version.__version, #major.minor[.patch[.sub]]
        description='Light-weight Python Computational Pipeline Management',
        maintainer="Leo Goodstadt",
        maintainer_email="ruffus_lib@llew.org.uk",
        author='Leo Goodstadt',
        author_email='ruffus@llew.org.uk',
        long_description=\
"""
***************************************
Overview
***************************************


    The Ruffus module is a lightweight way to add support
    for running computational pipelines.

    Computational pipelines are often conceptually quite simple, especially
    if we breakdown the process into simple stages, or separate **tasks**.

    Each stage or **task** in a computational pipeline is represented by a python function
    Each python function can be called in parallel to run multiple **jobs**.

    Ruffus was originally designed for use in bioinformatics to analyse multiple genome
    data sets.

***************************************
Documentation
***************************************

    Ruffus documentation can be found `here <http://www.ruffus.org.uk>`__ ,
    with `download notes <http://www.ruffus.org.uk/installation.html>`__ ,
    a `tutorial <http://www.ruffus.org.uk/tutorials/new_tutorial/introduction.html>`__ and
    an `in-depth manual <http://www.ruffus.org.uk/tutorials/new_tutorial/manual_contents.html>`__ .


***************************************
Background
***************************************

    The purpose of a pipeline is to determine automatically which parts of a multi-stage
    process needs to be run and in what order in order to reach an objective ("targets")

    Computational pipelines, especially for analysing large scientific datasets are
    in widespread use.
    However, even a conceptually simple series of steps can be difficult to set up and
    maintain.

***************************************
Design
***************************************
    The ruffus module has the following design goals:

        * Lightweight
        * Scalable / Flexible / Powerful
        * Standard Python
        * Unintrusive
        * As simple as possible

***************************************
Features
***************************************

    Automatic support for

        * Managing dependencies
        * Parallel jobs, including dispatching work to computational clusters
        * Re-starting from arbitrary points, especially after errors (checkpointing)
        * Display of the pipeline as a flowchart
        * Managing complex pipeline topologies


***************************************
A Simple example
***************************************

        Use the **@follows(...)** python decorator before the function definitions::

            from ruffus import *
            import sys

            def first_task():
                print "First task"

            @follows(first_task)
            def second_task():
                print "Second task"

            @follows(second_task)
            def final_task():
                print "Final task"




        the ``@follows`` decorator indicate that the ``first_task`` function precedes ``second_task`` in
        the pipeline.

        The canonical Ruffus decorator is ``@transform`` which **transforms** data flowing down a
        computational pipeline from one stage to teh next.

********
Usage
********

    Each stage or **task** in a computational pipeline is represented by a python function
    Each python function can be called in parallel to run multiple **jobs**.

    1. Import module::

            import ruffus


    1. Annotate functions with python decorators

    2. Print dependency graph if you necessary

        - For a graphical flowchart in ``jpg``, ``svg``, ``dot``, ``png``, ``ps``, ``gif`` formats::

            pipeline_printout_graph ("flowchart.svg")

        This requires ``dot`` to be installed

        - For a text printout of all jobs ::

            pipeline_printout(sys.stdout)


    3. Run the pipeline::

        pipeline_run()


""",
        url='http://www.ruffus.org.uk',
        download_url = "https://pypi.python.org/pypi/ruffus",

        install_requires = module_dependencies, #['multiprocessing>=1.0', 'json' ], #, 'python>=2.5'],
        setup_requires   = module_dependencies, #['multiprocessing>=1.0', 'json'],    #, 'python>=2.5'],


        classifiers=[
                    'Intended Audience :: End Users/Desktop',
                    'Development Status :: 5 - Production/Stable',
                    'Intended Audience :: Developers',
                    'Intended Audience :: Science/Research',
                    'Intended Audience :: Information Technology',
                    'License :: OSI Approved :: MIT License',
                    'Programming Language :: Python',
                    'Topic :: Scientific/Engineering',
                    'Topic :: Scientific/Engineering :: Bio-Informatics',
                    'Topic :: System :: Distributed Computing',
                    'Topic :: Software Development :: Build Tools',
                    'Topic :: Software Development :: Build Tools',
                    'Topic :: Software Development :: Libraries',
                    'Environment :: Console',
                    ],
        license = "MIT",
        keywords = "make task pipeline parallel bioinformatics science",


        #packages = find_packages('src'),    # include all packages under src
        #package_dir = {'':'src'},           #packages are under src
        packages=['ruffus'],
        package_dir={'ruffus': 'ruffus'},
        include_package_data = True,    # include everything in source control
        #package_data = {
        #    # If any package contains *.txt files, include them:
        #    '': ['*.TXT'],                                \
        #}


     )

#
#  http://pypi.python.org/pypi
#  http://docs.python.org/distutils/packageindex.html
#
#
#
# python setup.py register
# python setup.py sdist --format=gztar upload



#!/usr/bin/env python
from __future__ import print_function
import shutil
import unittest
import random
from ruffus.combinatorics import *
from ruffus.ruffus_utility import CHECKSUM_FILE_TIMESTAMPS
from ruffus.ruffus_exceptions import RethrownJobError
from ruffus import pipeline_run, pipeline_printout, Pipeline, follows, merge, posttask, \
    split, collate, mkdir, regex, files, files_re, combine
import sys

"""

    branching.py

        test branching dependencies

        use :
            --debug               to test automatically
            --start_again         the first time you run the file
            --jobs_per_task N     to simulate tasks with N numbers of files per task

            -j N / --jobs N       to speify multitasking
            -v                    to see the jobs in action
            -n / --just_print     to see what jobs would run

"""


import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   constants


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

JOBS_PER_TASK = 50
verbose_output = sys.stderr


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

# use simplejson in place of json for python < 2.6
# try:
#    import json
# except ImportError:
#    import simplejson
#    json = simplejson


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    split_fasta_file
#
@posttask(lambda: verbose_output.write("    Split into %d files\n" % JOBS_PER_TASK))
@files(tempdir + "original.fa", tempdir + "files.split.success")
def split_fasta_file(input_file, success_flag):
    #
    # remove existing fasta files
    #
    import glob
    filenames = sorted(glob.glob(tempdir + "files.split.*.fa"))
    for f in filenames:
        os.unlink(f)

    random.seed()
    for i in range(JOBS_PER_TASK):
        with open(tempdir + "files.split.%03d.fa" % i, "w") as oo:
            pass

    with open(success_flag,  "w") as oo:
        pass


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    align_sequences
#
@posttask(lambda: verbose_output.write("    Sequences aligned\n"))
@follows(split_fasta_file)
@files_re(tempdir + "files.split.*.fa",       # find all .fa files
          ".fa$", ".aln")                     # fa -> aln
def align_sequences(input_file, output_filename):
    with open(output_filename, "w") as oo:
        oo.write("%s\n" % output_filename)


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    percentage_identity
#
@posttask(lambda: verbose_output.write("    %Identity calculated\n"))
@files_re(align_sequences,                     # find all results from align_sequences
          # match file name root and substitute
          r"(.*\.)(.+).aln$",
          r'\g<0>',  # the original file
          [r"\1\2.pcid",  # .pcid suffix for the result
           r"\1\2.pcid_success"],  # .pcid_success to indicate job completed
          r"\2")  # extra parameter to remember the file index
def percentage_identity(input_file, output_files, split_index):
    (output_filename, success_flag_filename) = output_files
    with open(output_filename, "w") as oo:
        oo.write("%s\n" % split_index)
    with open(success_flag_filename, "w") as oo:
        pass


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    combine_results
#
@posttask(lambda: verbose_output.write("    Results recombined\n"))
@files_re(percentage_identity, combine(r".*.pcid$"),
          [tempdir + "all.combine_results",
           tempdir + "all.combine_results_success"])
def combine_results(input_files, output_files):
    """
    Combine all
    """
    (output_filename, success_flag_filename) = output_files
    out = open(output_filename, "w")
    for inp, flag in input_files:
        with open(inp) as ii:
            out.write(ii.read())
    out.close()
    with open(success_flag_filename, "w") as oo:
        pass


class Test_ruffus(unittest.TestCase):

    def tearDown(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass

    def setUp(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass
        os.makedirs(tempdir)
        with open(tempdir + "original.fa", "w") as oo:
            pass

    def test_ruffus(self):
        pipeline_run(multiprocess=100, verbose=0, pipeline="main")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
from ruffus import pipeline_run, pipeline_printout, Pipeline, formatter, transform, mkdir, originate
import unittest
import shutil
from ruffus.task import t_stream_logger
from ruffus.ruffus_exceptions import RethrownJobError
import ruffus
import re
import sys
import os

"""

    test_transform_formatter.py

"""
JOBS_PER_TASK = 5

tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0]))

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


try:
    from StringIO import StringIO
except:
    from io import StringIO


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
@mkdir(tempdir)
@originate([os.path.join(tempdir, ff + ".tmp") for ff in "abcd"])
def generate_initial_files(out_name):
    with open(out_name, 'w') as outfile:
        pass


@transform(input=generate_initial_files, filter=formatter(), output="{path[0]}/{basename[0]}.task1.{whatever}",
           extras=['echo {dynamic_message} > {some_file}'])
def transform_with_missing_formatter_args(input_file, output_files, output1):
    print("input = %r, output = %r, extras = %r" %
          (input_file, output_files, output1))


class Test_ruffus(unittest.TestCase):
    # ___________________________________________________________________________
    #
    #   setup and cleanup
    # ___________________________________________________________________________
    def setUp(self):
        import os
        try:
            shutil.rmtree(tempdir)
        except:
            pass
        os.makedirs(tempdir)

    def tearDown(self):
        shutil.rmtree(tempdir)

    # ___________________________________________________________________________
    #
    #   test product() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_transform_with_missing_formatter_args(self):
        s = StringIO()
        pipeline_printout(s, [transform_with_missing_formatter_args],
                          verbose=4, wrap_width=10000, pipeline="main")
        self.assertIn("Unmatched field {dynamic_message}", s.getvalue())
        pipeline_run([transform_with_missing_formatter_args],
                     verbose=0, pipeline="main")

    def test_transform_with_missing_formatter_args_b(self):
        test_pipeline = Pipeline("test")

        test_pipeline.originate(task_func=generate_initial_files,
                                output=[os.path.join(tempdir, ff + ".tmp") for ff in "abcd"])\
            .mkdir(tempdir)

        test_pipeline.transform(task_func=transform_with_missing_formatter_args,
                                input=generate_initial_files,
                                filter=formatter(),
                                output="{path[0]}/{basename[0]}.task1",
                                extras=['echo {dynamic_message} > {some_file}'])
        s = StringIO()
        test_pipeline.printout(
            s, [transform_with_missing_formatter_args], verbose=4, wrap_width=10000, pipeline="test")
        self.assertIn("Unmatched field {dynamic_message}", s.getvalue())

        # log to stream
        s = StringIO()
        logger = t_stream_logger(s)
        test_pipeline.run([transform_with_missing_formatter_args],
                          verbose=5, pipeline="test", logger=logger)
        self.assertIn("Unmatched field {dynamic_message}", s.getvalue())


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
from ruffus.ruffus_exceptions import JobSignalledBreak
from ruffus import transform, graphviz, check_if_uptodate, follows, Pipeline, \
    pipeline_run, pipeline_printout, pipeline_printout_graph
import sys

"""
    test_graphviz.py

"""
import unittest

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

try:
    from StringIO import StringIO
except:
    from io import StringIO, BytesIO


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Pipeline


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


#
# up to date tasks
#
@check_if_uptodate(lambda: (False, ""))
def Up_to_date_task1(infile, outfile):
    pass


@graphviz(URL='"http://cnn.com"', fillcolor='"#FFCCCC"',
          color='"#FF0000"', pencolor='"#FF0000"', fontcolor='"#4B6000"',
          label_suffix="???", label_prefix="What is this?<BR/> ",
          label="<What <FONT COLOR=\"red\">is</FONT>this>",
          shape="component", height=1.5, peripheries=5,
          style="dashed")
@check_if_uptodate(lambda: (False, ""))
@follows(Up_to_date_task1)
def Up_to_date_task2(infile, outfile):
    pass


@check_if_uptodate(lambda: (False, ""))
@follows(Up_to_date_task2)
def Up_to_date_task3(infile, outfile):
    pass


@check_if_uptodate(lambda: (False, ""))
@follows(Up_to_date_task3)
def Up_to_date_final_target(infile, outfile):
    pass


#
# Explicitly specified
#
@check_if_uptodate(lambda: (False, ""))
@follows(Up_to_date_task1)
def Explicitly_specified_task(infile, outfile):
    pass


#
# Tasks to run
#
@follows(Explicitly_specified_task)
def Task_to_run1(infile, outfile):
    pass


@follows(Task_to_run1)
def Task_to_run2(infile, outfile):
    pass


@follows(Task_to_run2)
def Task_to_run3(infile, outfile):
    pass


@check_if_uptodate(lambda: (False, ""))
@follows(Task_to_run2)
def Up_to_date_task_forced_to_rerun(infile, outfile):
    pass


#
# Final target
#
@follows(Up_to_date_task_forced_to_rerun, Task_to_run3)
def Final_target(infile, outfile):
    pass

#
# Ignored downstream
#


@follows(Final_target)
def Downstream_task1_ignored(infile, outfile):
    pass


@follows(Final_target)
def Downstream_task2_ignored(infile, outfile):
    pass


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


class Test_graphviz(unittest.TestCase):
    # ___________________________________________________________________________
    #
    #   test product() pipeline_printout diagnostic error messsages
    #
    #       require verbose >= 3 or an empty jobs list
    # ___________________________________________________________________________
    def test_graphviz_dot(self):
        """Make sure annotations from graphviz appear in dot
        """

        if sys.hexversion >= 0x03000000:
            # everything is unicode in python3
            s = BytesIO()
        else:
            s = StringIO()

        pipeline_printout_graph(
            s,
            # use flowchart file name extension to decide flowchart format
            #   e.g. svg, jpg etc.
            "dot",
            [Final_target, Up_to_date_final_target], pipeline="main")
        self.assertTrue('[URL="http://cnn.com", color="#FF0000", fillcolor="#FFCCCC", fontcolor="#4B6000", height=1.5, label=<What is this?<BR/> What <FONT COLOR="red">is</FONT>this???>, pencolor="#FF0000", peripheries=5, shape=component, style=dashed]' in s.getvalue().decode())

    def test_newstyle_graphviz_dot(self):
        test_pipeline = Pipeline("test")
        test_pipeline.check_if_uptodate(Up_to_date_task1, lambda: (False, ""))
        test_pipeline.follows(Up_to_date_task2, Up_to_date_task1)\
            .check_if_uptodate(lambda: (False, ""))\
            .graphviz(URL='"http://cnn.com"', fillcolor='"#FFCCCC"',
                      color='"#FF0000"', pencolor='"#FF0000"', fontcolor='"#4B6000"',
                      label_suffix="???", label_prefix="What is this?<BR/> ",
                      label="<What <FONT COLOR=\"red\">is</FONT>this>",
                      shape="component", height=1.5, peripheries=5,
                      style="dashed")
        test_pipeline.follows(Up_to_date_task3, Up_to_date_task2)\
            .check_if_uptodate(lambda: (False, ""))
        test_pipeline.follows(Up_to_date_final_target, Up_to_date_task3)\
            .check_if_uptodate(lambda: (False, ""))
        test_pipeline.follows(Explicitly_specified_task, Up_to_date_task1)\
            .check_if_uptodate(lambda: (False, ""))
        test_pipeline.follows(Task_to_run1, Explicitly_specified_task)
        test_pipeline.follows(Task_to_run2, Task_to_run1)
        test_pipeline.follows(Task_to_run3, Task_to_run2)
        test_pipeline.follows(Up_to_date_task_forced_to_rerun, Task_to_run2)\
            .check_if_uptodate(lambda: (False, ""))
        test_pipeline.follows(
            Final_target, Up_to_date_task_forced_to_rerun, Task_to_run3)
        test_pipeline.follows(Downstream_task1_ignored, Final_target)
        test_pipeline.follows(Downstream_task2_ignored, Final_target)

        if sys.hexversion >= 0x03000000:
            # everything is unicode in python3
            s = BytesIO()
        else:
            s = StringIO()

        test_pipeline.printout_graph(
            s,
            # use flowchart file name extension to decide flowchart format
            #   e.g. svg, jpg etc.
            "dot",
            [Final_target, Up_to_date_final_target])
        self.assertTrue('[URL="http://cnn.com", color="#FF0000", fillcolor="#FFCCCC", fontcolor="#4B6000", height=1.5, label=<What is this?<BR/> What <FONT COLOR="red">is</FONT>this???>, pencolor="#FF0000", peripheries=5, shape=component, style=dashed]' in s.getvalue().decode())


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    unittest.main()



import contextlib
import random
import unittest
import ruffus
import ruffus.drmaa_wrapper
import os
import shutil
import glob
import tempfile

try:
    import gevent
    HAVE_GEVENT = True
except ImportError:
    HAVE_GEVENT = False

try:
    import drmaa
    HAVE_DRMAA = True
    DRMAA_SESSION = drmaa.Session()
    DRMAA_SESSION.initialize()

except ImportError:
    HAVE_DRMAA = False


ROOT = os.path.abspath(os.path.dirname(__file__))
TESTS_TEMPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "tmp"))

# number of cores used for testing parallelism
NUM_CORES = 4


@contextlib.contextmanager
def temp_cd(target):
    curdir = os.path.curdir
    os.chdir(target)
    yield
    os.chdir(curdir)


def save_pid(outfile):
    with open(outfile + ".pid", "w") as outf:
        outf.write("{}".format(os.getpid()))


def create_files(outfile):
    with open(outfile, "w") as outf:
        outf.write("\n".join(map(
            str,
            random.sample(range(0, 1000), 100))) + "\n")
    save_pid(outfile)


def compute_mean(infile, outfile):
    """compute mean"""
    with open(infile, "r") as inf:
        n = [float(x.strip()) for x in inf.readlines()]

    with open(outfile, "w") as outf:
        outf.write("{}\n".format(sum(n) / len(n)))
    save_pid(outfile)


def run_job(infile, outfile, **kwargs):
    if not kwargs.get("run_locally", False):
        kwargs["drmaa_session"] = DRMAA_SESSION

    stdout, stderr = ruffus.drmaa_wrapper.run_job(
        cmd_str="hostname > {}".format(os.path.abspath(outfile)),
        verbose=1,
        local_echo=False,
        **kwargs)


def run_local_job1(*args):
    return run_job(*args, run_locally=True)


def run_local_job2(*args):
    return run_job(*args, run_locally=True)


def run_remote_job1(*args):
    return run_job(*args, run_locally=False)


def run_remote_job2(*args):
    return run_job(*args, run_locally=False)


def combine_means(infiles, outfile):
    with open(outfile, "w") as outf:
        for infile in infiles:
            with open(infile) as inf:
                outf.write(inf.read())
    save_pid(outfile)


class BaseTest(unittest.TestCase):

    expected_output_files = ["sample_{:02}.mean".format(x) for x in range(10)] +\
                            ["sample_{:02}.txt".format(x) for x in range(10)]

    def setUp(self):
        try:
            os.makedirs(TESTS_TEMPDIR)
        except OSError:
            pass
        self.work_dir = tempfile.mkdtemp(suffix="",
                                         prefix="ruffus_tmp_{}_".format(
                                             self.id()),
                                         dir=TESTS_TEMPDIR)

    def tearDown(self):
        # shutil.rmtree(self.work_dir)
        pass

    def check_files(self, present=[], absent=[]):
        for fn in present:
            path = os.path.join(self.work_dir, fn)
            self.assertTrue(os.path.exists(path),
                            "file {} does not exist".format(path))
        for fn in absent:
            path = os.path.join(self.work_dir, fn)
            self.assertFalse(os.path.exists(path),
                             "file {} does exist but not expected".format(path))

    def build_pipeline(self, pipeline_name, **kwargs):
        # fudge: clear all previous pipelines
        ruffus.Pipeline.clear_all()
        pipeline = ruffus.Pipeline(pipeline_name)

        task_create_files = pipeline.originate(
            task_func=create_files,
            output=["sample_{:02}.txt".format(x) for x in range(10)])

        task_compute_mean = pipeline.transform(
            task_func=compute_mean,
            input=task_create_files,
            filter=ruffus.suffix(".txt"),
            output=".mean")

        task_combine_means = pipeline.merge(
            task_func=combine_means,
            input=task_compute_mean,
            output="means.txt")

        task_run_local_job1 = pipeline.transform(
            task_func=run_local_job1,
            input=task_create_files,
            filter=ruffus.suffix(".txt"),
            output=".local1")

        # test jobs_limit with local running
        task_run_local_job2 = pipeline.transform(
            task_func=run_local_job2,
            input=task_create_files,
            filter=ruffus.suffix(".txt"),
            output=".local2").jobs_limit(NUM_CORES // 2)

        # multiprocessing and DRMAA do not work at the moment likely
        # cause is the shared session object.
        if not HAVE_DRMAA or (kwargs.get("multiprocess", 1) > 1):
            return

        task_run_remote_job1 = pipeline.transform(
            task_func=run_remote_job1,
            input=task_create_files,
            filter=ruffus.suffix(".txt"),
            output=".remote1")

        # test jobs_limit with remote running
        task_run_remote_job2 = pipeline.transform(
            task_func=run_remote_job2,
            input=task_create_files,
            filter=ruffus.suffix(".txt"),
            output=".remote2").jobs_limit(NUM_CORES // 2)

    def run_pipeline(self, **kwargs):
        pipeline = self.build_pipeline(self.id(), **kwargs)
        with temp_cd(self.work_dir):
            ruffus.pipeline_run(pipeline=pipeline, verbose=5, **kwargs)
        self.check_files(self.expected_output_files)

    def read_pids(self):
        pids = []
        for fn in glob.glob(os.path.join(self.work_dir, "*.pid")):
            with open(fn) as inf:
                pids.append(int(inf.readline().strip()))
        return pids


class TestExecutionEngines(BaseTest):

    def test_pipeline_runs_with_multiprocessing(self):
        self.run_pipeline(multiprocess=NUM_CORES)
        pids = self.read_pids()
        self.assertEqual(len(set(pids)), NUM_CORES)

    def test_pipeline_runs_with_multithreading(self):
        self.run_pipeline(multithread=NUM_CORES)
        pids = self.read_pids()
        self.assertEqual(len(set(pids)), 1)
        self.assertEqual(pids[0], os.getpid())

    @unittest.skipIf(not HAVE_GEVENT, "no gevent installed")
    def test_pipeline_runs_with_gevent_manager(self):
        self.run_pipeline(multithread=NUM_CORES, pool_manager="gevent")
        pids = self.read_pids()
        self.assertEqual(len(set(pids)), 1)
        self.assertEqual(pids[0], os.getpid())

    def test_pipeline_fails_with_unknown_manager(self):
        self.assertRaises(ValueError,
                          self.run_pipeline,
                          multithread=NUM_CORES,
                          pool_manager="mystical_manager")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import re
import shutil
import unittest
from ruffus.combinatorics import *
from ruffus.ruffus_utility import RUFFUS_HISTORY_FILE, CHECKSUM_FILE_TIMESTAMPS
from ruffus.ruffus_exceptions import RethrownJobError
from ruffus import pipeline_run, pipeline_printout, Pipeline, formatter, originate, follows, merge
import ruffus
import sys

"""

    test_combinatorics.py

        test product, combine, permute, combine_with_replacement

"""

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0]))

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


try:
    from StringIO import StringIO
except:
    from io import StringIO


# sub-1s resolution in system?
one_second_per_job = None


def touch(filename):
    with open(filename, "w"):
        pass


# ___________________________________________________________________________
#
#   generate_initial_files1
# ___________________________________________________________________________
@originate(output=[tempdir + "/" + prefix + "_name.tmp1" for prefix in "abcd"])
def generate_initial_files1(out_name):
    with open(out_name, 'w') as outfile:
        pass

# ___________________________________________________________________________
#
#   generate_initial_files1
# ___________________________________________________________________________


@originate([tempdir + "/e_name.tmp1", tempdir + "/f_name.tmp1"])
def generate_initial_files2(out_name):
    with open(out_name, 'w') as outfile:
        pass

# ___________________________________________________________________________
#
#   generate_initial_files1
# ___________________________________________________________________________


@originate([tempdir + "/g_name.tmp1", tempdir + "/h_name.tmp1"])
def generate_initial_files3(out_name):
    with open(out_name, 'w') as outfile:
        pass

# ___________________________________________________________________________
#
#   check_product_task
# ___________________________________________________________________________


@follows(generate_initial_files1)
@product(
    [tempdir + "/" + prefix + "_name.tmp1" for prefix in "abcd"],
    formatter(".*/(?P<FILE_PART>.+).tmp1$"),
    generate_initial_files2,
    formatter(),
    "{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.{basename[2][0]}.tmp2",
    input3=generate_initial_files3,
    filter3=formatter(r"tmp1$"),
    extras=["{basename[0][0][0]}{basename[1][0][0]}{basename[2][0][0]}",       # extra: prefices only (abcd etc)
            # extra: path for 2nd input, 1st file
            "{subpath[0][0][0]}",
            "{subdir[0][0][0]}"])
def check_product_task(infiles, outfile,
                       prefices,
                       subpath,
                       subdir):
    with open(outfile, "w") as p:
        p.write(prefices + ",")


# ___________________________________________________________________________
#
#   check_product_merged_task
# ___________________________________________________________________________
@merge(check_product_task, tempdir + "/merged.results")
def check_product_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())

# ___________________________________________________________________________
#
#   check_product_misspelt_capture_error_task
# ___________________________________________________________________________


@product(
    generate_initial_files1,
    formatter(".*/(?P<FILE_PART>.+).tmp1$"),
    "{path[0][0]}/{FILEPART[0][0]}.tmp2")
def check_product_misspelt_capture_error_task(infiles, outfile):
    """
    FILE_PART mispelt as FILE_PART
    """
    with open(outfile, "w") as p:
        pass


# ___________________________________________________________________________
#
#   check_product_out_of_range_formatter_ref_error_task
# ___________________________________________________________________________
@product(
    generate_initial_files1,
    formatter(".*/(?P<FILE_PART>.+).tmp1$"),
    "{path[2][0]}/{basename[0][0]}.tmp2",
    "{FILE_PART[0][0]}")
def check_product_out_of_range_formatter_ref_error_task(infiles, outfile, ignored_filter):
    """
    {path[2][0]} when len(path) == 1
    """
    with open(outfile, "w") as p:
        pass

# ___________________________________________________________________________
#
#   check_product_formatter_ref_index_error_task
# ___________________________________________________________________________


@product(
    generate_initial_files1,
    formatter(".*/(?P<FILE_PART>.+).tmp1$"),
    "{path[0][0][1000]}/{basename[0][0]}.tmp2",
    "{FILE_PART[0][0]}")
def check_product_formatter_ref_index_error_task(infiles, outfile, ignored_filter):
    """
    {path[0][0][1000} when len of the path string len(path[0][0]) < 1000
    """
    with open(outfile, "w") as p:
        pass

# ___________________________________________________________________________
#
#   check_combinations2_task
# ___________________________________________________________________________


@combinations(
    generate_initial_files1,
    formatter(".*/(?P<FILE_PART>.+).tmp1$"),
    2,
    "{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.tmp2",
    "{basename[0][0][0]}{basename[1][0][0]}",       # extra: prefices
    "{subpath[0][0][0]}",      # extra: path for 2nd input, 1st file
    "{subdir[0][0][0]}")
def check_combinations2_task(infiles, outfile,
                             prefices,
                             subpath,
                             subdir):
    """
        Test combinations with k-tuple = 2
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


@merge(check_combinations2_task, tempdir + "/merged.results")
def check_combinations2_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())

# ___________________________________________________________________________
#
#   check_combinations3_task
# ___________________________________________________________________________


@combinations(
    generate_initial_files1,
    formatter(".*/(?P<FILE_PART>.+).tmp1$"),
    3,
    "{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.{basename[2][0]}.tmp2",
    # extra: prefices
    "{basename[0][0][0]}{basename[1][0][0]}{basename[2][0][0]}",
    "{subpath[0][0][0]}",      # extra: path for 2nd input, 1st file
    "{subdir[0][0][0]}")
def check_combinations3_task(infiles, outfile,
                             prefices,
                             subpath,
                             subdir):
    """
        Test combinations with k-tuple = 3
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


@merge(check_combinations3_task, tempdir + "/merged.results")
def check_combinations3_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())


# ___________________________________________________________________________
#
#   check_permutations2_task
# ___________________________________________________________________________
@permutations(
    generate_initial_files1,
    formatter(".*/(?P<FILE_PART>.+).tmp1$"),
    2,
    "{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.tmp2",
    "{basename[0][0][0]}{basename[1][0][0]}",       # extra: prefices
    "{subpath[0][0][0]}",      # extra: path for 2nd input, 1st file
    "{subdir[0][0][0]}")
def check_permutations2_task(infiles, outfile,
                             prefices,
                             subpath,
                             subdir):
    """
        Test permutations with k-tuple = 2
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


@merge(check_permutations2_task, tempdir + "/merged.results")
def check_permutations2_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())


# ___________________________________________________________________________
#
#   check_permutations3_task
# ___________________________________________________________________________
@permutations(
    generate_initial_files1,
    formatter(".*/(?P<FILE_PART>.+).tmp1$"),
    3,
    "{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.{basename[2][0]}.tmp2",
    # extra: prefices
    "{basename[0][0][0]}{basename[1][0][0]}{basename[2][0][0]}",
    "{subpath[0][0][0]}",      # extra: path for 2nd input, 1st file
    "{subdir[0][0][0]}")
def check_permutations3_task(infiles, outfile,
                             prefices,
                             subpath,
                             subdir):
    """
        Test permutations with k-tuple = 3
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


@merge(check_permutations3_task, tempdir + "/merged.results")
def check_permutations3_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())


# ___________________________________________________________________________
#
#   check_combinations_with_replacement2_task
# ___________________________________________________________________________
@combinations_with_replacement(
    input=generate_initial_files1,
    filter=formatter(".*/(?P<FILE_PART>.+).tmp1$"),
    tuple_size=2,
    output="{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.tmp2",
    extras=["{basename[0][0][0]}{basename[1][0][0]}",       # extra: prefices
            "{subpath[0][0][0]}",      # extra: path for 2nd input, 1st file
            "{subdir[0][0][0]}"])
def check_combinations_with_replacement2_task(infiles, outfile,
                                              prefices,
                                              subpath,
                                              subdir):
    """
        Test combinations_with_replacement with k-tuple = 2
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


@merge(check_combinations_with_replacement2_task, tempdir + "/merged.results")
def check_combinations_with_replacement2_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())


# ___________________________________________________________________________
#
#   check_combinations_with_replacement3_task
# ___________________________________________________________________________
@combinations_with_replacement(
    generate_initial_files1,
    formatter(".*/(?P<FILE_PART>.+).tmp1$"),
    3,
    "{path[0][0]}/{FILE_PART[0][0]}.{basename[1][0]}.{basename[2][0]}.tmp2",
    # extra: prefices
    "{basename[0][0][0]}{basename[1][0][0]}{basename[2][0][0]}",
    "{subpath[0][0][0]}",      # extra: path for 2nd input, 1st file
    "{subdir[0][0][0]}")
def check_combinations_with_replacement3_task(infiles, outfile,
                                              prefices,
                                              subpath,
                                              subdir):
    """
        Test combinations_with_replacement with k-tuple = 3
    """
    with open(outfile, "w") as outf:
        outf.write(prefices + ",")


@merge(check_combinations_with_replacement3_task, tempdir + "/merged.results")
def check_combinations_with_replacement3_merged_task(infiles, outfile):
    with open(outfile, "w") as p:
        for infile in sorted(infiles):
            with open(infile) as ii:
                p.write(ii.read())


def cleanup_tmpdir():
    os.system('rm -f %s %s' %
              (os.path.join(tempdir, '*'), RUFFUS_HISTORY_FILE))


class TestCombinatorics(unittest.TestCase):
    def setUp(self):
        try:
            os.mkdir(tempdir)
        except OSError:
            pass

    # ___________________________________________________________________________
    #
    #   test product() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_product_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()
        s = StringIO()
        pipeline_printout(s, [check_product_merged_task],
                          verbose=5, wrap_width=10000, pipeline="main")
        self.assertTrue(re.search('Job needs update:.*Missing files.*'
                                  '\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/e_name.tmp1, '
                                  '.*{tempdir}/h_name.tmp1, '
                                  '.*{tempdir}/a_name.e_name.h_name.tmp2\]'.format(tempdir=tempdir), s.getvalue(), re.DOTALL))

    def test_product_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        pipeline_run([check_product_merged_task], verbose=0, multiprocess=100,
                     one_second_per_job=one_second_per_job, pipeline="main")
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             "aeg,aeh,afg,afh,beg,beh,bfg,bfh,ceg,ceh,cfg,cfh,deg,deh,dfg,dfh,")

    # ___________________________________________________________________________
    #
    #   test product() pipeline_printout diagnostic error messsages
    #
    #       require verbose >= 3 or an empty jobs list
    # ___________________________________________________________________________
    def test_product_misspelt_capture_error(self):
        """Misspelt named capture group
            Requires verbose >= 3 or an empty jobs list
        """
        cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [check_product_misspelt_capture_error_task],
                          verbose=3, wrap_width=10000, pipeline="main")
        self.assertIn("Warning: Input substitution failed:", s.getvalue())
        self.assertIn("Unmatched field {FILEPART}", s.getvalue())

    def test_product_out_of_range_formatter_ref_error(self):
        """
        {path[2][0]} when len(path) == 1
            Requires verbose >= 3 or an empty jobs list
        """
        cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [check_product_out_of_range_formatter_ref_error_task],
                          verbose=3, wrap_width=10000, pipeline="main")
        self.assertIn("Warning: Input substitution failed:", s.getvalue())
        self.assertIn("Unmatched field {2}", s.getvalue())

    def test_product_formatter_ref_index_error(self):
        """
        {path[0][0][1000} when len of the path string len(path[0][0]) < 1000
            Requires verbose >= 3 or an empty jobs list
        """
        cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [check_product_formatter_ref_index_error_task],
                          verbose=3, wrap_width=10000, pipeline="main")
        self.assertIn("Warning: Input substitution failed:", s.getvalue())
        self.assertIn(
            "Unmatched field {string index out of range}", s.getvalue())

        # print s.getvalue()

    # ___________________________________________________________________________
    #
    #   test combinations() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_combinations2_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [check_combinations2_merged_task],
                          verbose=5, wrap_width=10000, pipeline="main")
        self.assertTrue(re.search('Job needs update:.*Missing files.*'
                                  '\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/b_name.tmp1, '
                                  '.*{tempdir}/a_name.b_name.tmp2\]'.format(tempdir=tempdir), s.getvalue(), re.DOTALL))

    def test_combinations2_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        pipeline_run([check_combinations2_merged_task], verbose=0, multiprocess=100,
                     one_second_per_job=one_second_per_job, pipeline="main")
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             'ab,ac,ad,bc,bd,cd,')

    # ___________________________________________________________________________
    #
    #   test combinations() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_combinations3_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [check_combinations3_merged_task],
                          verbose=5, wrap_width=10000, pipeline="main")
        self.assertTrue(re.search(
            '\[.*{tempdir}/a_name.tmp1, '
            '.*{tempdir}/b_name.tmp1, '
            '.*{tempdir}/c_name.tmp1, '
            '.*{tempdir}/a_name.b_name.c_name.tmp2\]'.format(tempdir=tempdir), s.getvalue()))

    def test_combinations3_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        pipeline_run([check_combinations3_merged_task], verbose=0, multiprocess=100,
                     one_second_per_job=one_second_per_job, pipeline="main")
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             "abc,abd,acd,bcd,")

    # ___________________________________________________________________________
    #
    #   test permutations() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_permutations2_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [check_permutations2_merged_task],
                          verbose=5, wrap_width=10000, pipeline="main")
        self.assertTrue(re.search('\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/b_name.tmp1, '
                                  '.*{tempdir}/a_name.b_name.tmp2\]'.format(tempdir=tempdir), s.getvalue()))

    def test_permutations2_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        pipeline_run([check_permutations2_merged_task], verbose=0, multiprocess=100,
                     one_second_per_job=one_second_per_job, pipeline="main")
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             "ab,ac,ad,ba,bc,bd,ca,cb,cd,da,db,dc,")

    # ___________________________________________________________________________
    #
    #   test permutations() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_permutations3_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [check_permutations3_merged_task],
                          verbose=5, wrap_width=10000, pipeline="main")
        self.assertTrue(re.search('\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/b_name.tmp1, '
                                  '.*{tempdir}/c_name.tmp1, '
                                  '.*{tempdir}/a_name.b_name.c_name.tmp2\]'.format(tempdir=tempdir), s.getvalue()))

    def test_permutations3_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        pipeline_run([check_permutations3_merged_task], verbose=0, multiprocess=100,
                     one_second_per_job=one_second_per_job, pipeline="main")
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             'abc,abd,acb,acd,adb,adc,bac,bad,bca,bcd,bda,bdc,cab,cad,cba,cbd,cda,cdb,dab,dac,dba,dbc,dca,dcb,')

    # ___________________________________________________________________________
    #
    #   test combinations_with_replacement() pipeline_printout and pipeline_run
    # ___________________________________________________________________________

    def test_combinations_with_replacement2_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [check_combinations_with_replacement2_merged_task],
                          verbose=5, wrap_width=10000, pipeline="main")
        self.assertTrue(re.search('\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/b_name.tmp1, '
                                  '.*{tempdir}/a_name.b_name.tmp2\]'.format(tempdir=tempdir), s.getvalue()))

    def test_combinations_with_replacement2_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        pipeline_run([check_combinations_with_replacement2_merged_task], verbose=0,
                     multiprocess=100, one_second_per_job=one_second_per_job, pipeline="main")
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             "aa,ab,ac,ad,bb,bc,bd,cc,cd,dd,")

    # ___________________________________________________________________________
    #
    #   test combinations_with_replacement() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_combinations_with_replacement3_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [check_combinations_with_replacement3_merged_task],
                          verbose=5, wrap_width=10000, pipeline="main")
        self.assertTrue(re.search('\[.*{tempdir}/a_name.tmp1, '
                                  '.*{tempdir}/b_name.tmp1, '
                                  '.*{tempdir}/c_name.tmp1, '
                                  '.*{tempdir}/a_name.b_name.c_name.tmp2\]'.format(tempdir=tempdir), s.getvalue()))

    def test_combinations_with_replacement3_run(self):
        """Run product"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        pipeline_run([check_combinations_with_replacement3_merged_task], verbose=0,
                     multiprocess=100, one_second_per_job=one_second_per_job, pipeline="main")
        with open(tempdir + "/merged.results") as oo:
            self.assertEqual(oo.read(),
                             'aaa,aab,aac,aad,abb,abc,abd,acc,acd,add,bbb,bbc,bbd,bcc,bcd,bdd,ccc,ccd,cdd,ddd,')

    # ___________________________________________________________________________
    #
    #   cleanup
    # ___________________________________________________________________________

    def tearDown(self):
        shutil.rmtree(tempdir)


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    #pipeline_printout(sys.stdout, [check_product_task], verbose = 5, pipeline= "main")
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
from collections import defaultdict
from ruffus.ruffus_exceptions import JobSignalledBreak
from ruffus import *
from optparse import OptionParser
"""

    play_with_colours.py
    [--log_file PATH]
    [--verbose]

"""

################################################################################
#
#   test
#
#
#   Copyright (c) 7/13/2010 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################

import sys
import os

# add self to search path for testing
exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
sys.path.insert(0, os.path.abspath(os.path.join(exe_path, "..", "..")))
if __name__ == '__main__':
    module_name = os.path.split(sys.argv[0])[1]
    module_name = os.path.splitext(module_name)[0]
else:
    module_name = __name__


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


try:
    import StringIO as io
except:
    import io as io

parser = OptionParser(version="%play_with_colours 1.0",
                      usage="\n\n    play_with_colours "
                      "--flowchart FILE [options] "
                      "[--colour_scheme_index INT ] "
                      "[--key_legend_in_graph]")

#
#   pipeline
#
parser.add_option("--flowchart", dest="flowchart",
                  metavar="FILE",
                  type="string",
                  help="Don't actually run any commands; just print the pipeline "
                  "as a flowchart.")
parser.add_option("--colour_scheme_index", dest="colour_scheme_index",
                  metavar="INTEGER",
                  type="int",
                  help="Index of colour scheme for flow chart.")
parser.add_option("--key_legend_in_graph", dest="key_legend_in_graph",
                  action="store_true", default=False,
                  help="Print out legend and key for dependency graph.")

(options, remaining_args) = parser.parse_args()
if not options.flowchart:
    raise Exception("Missing mandatory parameter: --flowchart.\n")


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Pipeline


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


#
# up to date tasks
#
@check_if_uptodate(lambda: (False, ""))
def Up_to_date_task1(infile, outfile):
    pass


@check_if_uptodate(lambda: (False, ""))
@follows(Up_to_date_task1)
def Up_to_date_task2(infile, outfile):
    pass


@check_if_uptodate(lambda: (False, ""))
@follows(Up_to_date_task2)
def Up_to_date_task3(infile, outfile):
    pass


@check_if_uptodate(lambda: (False, ""))
@follows(Up_to_date_task3)
def Up_to_date_final_target(infile, outfile):
    pass


#
# Explicitly specified
#
@check_if_uptodate(lambda: (False, ""))
@follows(Up_to_date_task1)
def Explicitly_specified_task(infile, outfile):
    pass


#
# Tasks to run
#
@follows(Explicitly_specified_task)
def Task_to_run1(infile, outfile):
    pass


@follows(Task_to_run1)
def Task_to_run2(infile, outfile):
    pass


@follows(Task_to_run2)
def Task_to_run3(infile, outfile):
    pass


@check_if_uptodate(lambda: (False, ""))
@follows(Task_to_run2)
def Up_to_date_task_forced_to_rerun(infile, outfile):
    pass


#
# Final target
#
@follows(Up_to_date_task_forced_to_rerun, Task_to_run3)
def Final_target(infile, outfile):
    pass

#
# Ignored downstream
#


@follows(Final_target)
def Downstream_task1_ignored(infile, outfile):
    pass


@follows(Final_target)
def Downstream_task2_ignored(infile, outfile):
    pass


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#   Main logic
# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
custom_flow_chart_colour_scheme = defaultdict(dict)

#
#   Base chart on this overall colour scheme index
#
custom_flow_chart_colour_scheme["colour_scheme_index"] = options.colour_scheme_index

#
#   Overriding colours
#
if options.colour_scheme_index is None:
    custom_flow_chart_colour_scheme["Vicious cycle"]["linecolor"] = '"#FF3232"'
    custom_flow_chart_colour_scheme["Pipeline"]["fontcolor"] = '"#FF3232"'
    custom_flow_chart_colour_scheme["Key"]["fontcolor"] = "black"
    custom_flow_chart_colour_scheme["Key"]["fillcolor"] = '"#F6F4F4"'
    custom_flow_chart_colour_scheme["Task to run"]["linecolor"] = '"#0044A0"'
    custom_flow_chart_colour_scheme["Up-to-date"]["linecolor"] = "gray"
    custom_flow_chart_colour_scheme["Final target"]["fillcolor"] = '"#EFA03B"'
    custom_flow_chart_colour_scheme["Final target"]["fontcolor"] = "black"
    custom_flow_chart_colour_scheme["Final target"]["color"] = "black"
    custom_flow_chart_colour_scheme["Final target"]["dashed"] = 0
    custom_flow_chart_colour_scheme["Vicious cycle"]["fillcolor"] = '"#FF3232"'
    custom_flow_chart_colour_scheme["Vicious cycle"]["fontcolor"] = 'white'
    custom_flow_chart_colour_scheme["Vicious cycle"]["color"] = "white"
    custom_flow_chart_colour_scheme["Vicious cycle"]["dashed"] = 0
    custom_flow_chart_colour_scheme["Up-to-date task"]["fillcolor"] = '"#B8CC6E"'
    custom_flow_chart_colour_scheme["Up-to-date task"]["fontcolor"] = '"#006000"'
    custom_flow_chart_colour_scheme["Up-to-date task"]["color"] = '"#006000"'
    custom_flow_chart_colour_scheme["Up-to-date task"]["dashed"] = 0
    custom_flow_chart_colour_scheme["Down stream"]["fillcolor"] = "white"
    custom_flow_chart_colour_scheme["Down stream"]["fontcolor"] = "gray"
    custom_flow_chart_colour_scheme["Down stream"]["color"] = "gray"
    custom_flow_chart_colour_scheme["Down stream"]["dashed"] = 0
    custom_flow_chart_colour_scheme["Explicitly specified task"]["fillcolor"] = "transparent"
    custom_flow_chart_colour_scheme["Explicitly specified task"]["fontcolor"] = "black"
    custom_flow_chart_colour_scheme["Explicitly specified task"]["color"] = "black"
    custom_flow_chart_colour_scheme["Explicitly specified task"]["dashed"] = 0
    custom_flow_chart_colour_scheme["Task to run"]["fillcolor"] = '"#EBF3FF"'
    custom_flow_chart_colour_scheme["Task to run"]["fontcolor"] = '"#0044A0"'
    custom_flow_chart_colour_scheme["Task to run"]["color"] = '"#0044A0"'
    custom_flow_chart_colour_scheme["Task to run"]["dashed"] = 0
    custom_flow_chart_colour_scheme["Up-to-date task forced to rerun"]["fillcolor"] = 'transparent'
    custom_flow_chart_colour_scheme["Up-to-date task forced to rerun"]["fontcolor"] = '"#0044A0"'
    custom_flow_chart_colour_scheme["Up-to-date task forced to rerun"]["color"] = '"#0044A0"'
    custom_flow_chart_colour_scheme["Up-to-date task forced to rerun"]["dashed"] = 1
    custom_flow_chart_colour_scheme["Up-to-date Final target"]["fillcolor"] = '"#EFA03B"'
    custom_flow_chart_colour_scheme["Up-to-date Final target"]["fontcolor"] = '"#006000"'
    custom_flow_chart_colour_scheme["Up-to-date Final target"]["color"] = '"#006000"'
    custom_flow_chart_colour_scheme["Up-to-date Final target"]["dashed"] = 0

if __name__ == '__main__':
    pipeline_printout_graph(

        open(options.flowchart, "w"),
        # use flowchart file name extension to decide flowchart format
        #   e.g. svg, jpg etc.
        os.path.splitext(options.flowchart)[1][1:],

        # final targets
        [Final_target, Up_to_date_final_target],

        # Explicitly specified tasks
        [Explicitly_specified_task],

        # Do we want key legend
        no_key_legend=not options.key_legend_in_graph,

        # Print all the task types whether used or not
        minimal_key_legend=False,

        user_colour_scheme=custom_flow_chart_colour_scheme,
        pipeline_name="Colour schemes",
        pipeline="main")



#!/usr/bin/env python
"""

    complicated_example.py

"""

import os, sys
exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
sys.path.append(os.path.abspath(os.path.join(exe_path,"..", "..")))
from ruffus import *
from time import sleep
import random

#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options        


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

from optparse import OptionParser
parser = OptionParser(version="%prog 1.0")
parser.add_option("-t", "--target_tasks", dest="target_tasks",
                  action="append",
                  default = ["summarise_all"],
                  metavar="JOBNAME", 
                  type="string",
                  help="Target task(s) of pipeline.")
parser.add_option("-f", "--forced_tasks", dest="forced_tasks",
                  action="append",
                  default = list(),
                  metavar="JOBNAME", 
                  type="string",
                  help="Pipeline task(s) which will be included even if they are up to date.")
parser.add_option("-j", "--jobs", dest="jobs",
                  default=5,
                  metavar="jobs", 
                  type="int",
                  help="Specifies the number of jobs (commands) to run simultaneously.")

parser.add_option("-d", "--data_dir", dest="data_dir",
                  default="%s/data_for_complicated_example" % exe_path,
                  metavar="PATH", 
                  type="string",
                  help="Directory with starting data [*.fa].")
parser.add_option("-w", "--working_dir", dest="working_dir",
                  default="/working_dir",
                  metavar="PATH", 
                  type="string",
                  help="Working directory.")


parser.add_option("-v", "--verbose", dest = "verbose",
                  action="store_true", default=False,
                  help="Do not echo to shell but only print to log.")
parser.add_option("-D", "--dependency", dest="dependency_file",
                  metavar="FILE", 
                  type="string",
                  help="Print a dependency graph of the pipeline that would be executed "
                        "to FILE, but do not execute it.")
parser.add_option("-F", "--dependency_graph_format", dest="dependency_graph_format",
                  metavar="FORMAT", 
                  type="string",
                  default = 'svg',
                  help="format of dependency graph file. Can be 'ps' (PostScript), "+
                  "'svg' 'svgz' (Structured Vector Graphics), " +
                  "'png' 'gif' (bitmap  graphics) etc ")
parser.add_option("-n", "--just_print", dest="just_print",
                    action="store_true", default=False,
                    help="Print a description of the jobs that would be executed, "
                        "but do not execute them.")





#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports        


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

import StringIO
import re
import operator
import sys
from collections import defaultdict

#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Functions


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#_________________________________________________________________________________________
#
#   Helper function:
# 
#       split_gene_files 
#
#_________________________________________________________________________________________
def split_gene_files (  gene_file_name, 
                        job_completion_flag_file_name,
                        split_output_dir):
    """
    Helper function to simulate splitting gene files into "chunks" suitable for
        parallel jobs on a computational cluster       

    The number of output files is only known at runtime
        because the number of "chunks" depend on the size    
        of starting the gene sets                            

    We simulate this using a random number from 20->50 
    """

    #
    #   make output directory
    #    
    if not os.path.exists(split_output_dir):
        os.makedirs(split_output_dir)

    # save number of chunks for later tasks
    number_of_output_files = int(random.uniform(20, 50))

    for index in range(number_of_output_files):
        open("%s/%d.fa" % (split_output_dir, index), "w")
    open(job_completion_flag_file_name, "w")


#_________________________________________________________________________________________
#
#   get_unknown_gene_set_names
#   get_species_names
#   
#   
#   functions for getting unknown gene set names and species names
#
#_________________________________________________________________________________________
import glob, re
def get_chunked_gene_file_names (dir_name):
    """
    Get list of gene file names
    Helper function for getting unknown gene set names, and species names
    """
    regex = re.compile(r".+/(.+).genes.fa")
    gene_set_names = []
    for file_name in glob.glob("%s/%s/*.genes.fa"   % (d_dir, dir_name)):
        m = regex.search(file_name)
        gene_set_names.append(m.group(1))
    return gene_set_names
def get_unknown_gene_set_names ():
    return get_chunked_gene_file_names("unknown_genes")
def get_species_names ():
    return get_chunked_gene_file_names("all_genes_in_each_species")



#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888





# get help string
f =StringIO.StringIO()
parser.print_help(f)
helpstr = f.getvalue()
(options, remaining_args) = parser.parse_args()


d_dir = options.data_dir
w_dir = options.working_dir




#_________________________________________________________________________________________
#
#   Step 1:
#       
#       split_unknown_gene_set
#       
#           data_dir/unknown_genes/XXX.genes.fa
#               ->working_dir/XXX/split_gene_sets.completed
#               ->working_dir/XXX/NNN.fa
#       
#_________________________________________________________________________________________
@follows(mkdir(w_dir))
@files_re("%s/unknown_genes/*.genes.fa"      % d_dir,
          r"(.*/)(.*)(\.genes.fa)",                    
          r"\1\2\3",                                   # unknown_gene_set file name
          r"%s/\2/split_gene_sets.completed" % w_dir,  # job_completion_flag
          r"%s/\2"                           % w_dir)  # split_output_dir
def split_unknown_gene_set( starting_gene_set, 
                            job_completion_flag,
                            split_output_dir):
    """
    Simulate splitting gene files for unknown gene set into "chunks" suitable for
        parallel jobs on a computational cluster       
    """
    split_gene_files (  starting_gene_set, 
                        job_completion_flag,
                        split_output_dir)
    sleep(1)
    
    
#_________________________________________________________________________________________
#
#   Step 2:
#       
#       split_per_species_gene_sets

#           data_dir/all_genes_in_each_species/YYY.genes.fa
#               ->working_dir/species_YYY/split_gene_sets.completed
#               ->working_dir/species_YYY/MMM.fa
#       
#_________________________________________________________________________________________
@follows(mkdir(w_dir))
@files_re("%s/all_genes_in_each_species/*.genes.fa"  % d_dir,
          r"(.*/)(.*)(\.genes.fa)",                    
          r"\1\2\3",                                           # all_genes_in_species
          r"%s/species_\2/split_gene_sets.completed" % w_dir,  # job_completion_flag
          r"%s/species_\2"                           % w_dir)  # split_output_dir
def split_per_species_gene_sets(all_genes_in_species, 
                                job_completion_flag,
                                split_output_dir):
    """
    Simulate splitting gene files for each species into "chunks" suitable for
        parallel jobs on a computational cluster       
    """
    split_gene_files (  all_genes_in_species, 
                        job_completion_flag,
                        split_output_dir)
    sleep(1)



#_________________________________________________________________________________________
#
#   Step 3:
#       
#       all_vs_all_comparisons
#                 working_dir/species_YYY/MMM.fa
#                 working_dir/XXX/NNN.fa
#                                 -> compare/x/y.n.m.comparison_res
#                                 -> compare/x/y.n.m.complete
#       
#_________________________________________________________________________________________
#
#           function for generating custom parameters
#
def generate_all_vs_all_params ():
    """
    Custom function to generate 
    all vs. all file names for the various "chunks"
    """

    chunk_index_regex = re.compile(r".+/(.+).fa")
    def parse_index_from_chunk_filename (chunk_filename):
        match = chunk_index_regex.search(chunk_filename)
        return int(match.group(1))
       
    species_names = get_species_names()
    gene_set_names = get_unknown_gene_set_names()
    for x in gene_set_names:
        for y in species_names:
            y = "species_" + y
            
            m_files = glob.glob("%s/%s/*.fa" % (w_dir, x))
            n_files = glob.glob("%s/%s/*.fa" % (w_dir, y))

            #
            #    for each species chunk vs for each unknown chunk
            # 
            for m_file in m_files:
                for n_file in n_files:
                    input_files = [m_file, n_file]
                    output_dir  = "%s/compare/%s" % (w_dir, x)

                    m = parse_index_from_chunk_filename(m_file)
                    n = parse_index_from_chunk_filename(n_file)

                    job_completion_flag  = output_dir + "/%s.%d.%d.complete" % (y, m, n)
                    result_file          = output_dir + "/%s.%d.%d.comparison_res" % (y, m, n)
                    name                 = "%s -> %d vs %d\n" % (y, m, n)
                    yield input_files, job_completion_flag, output_dir, result_file, name
    

    
@follows(split_unknown_gene_set, split_per_species_gene_sets)
@files(generate_all_vs_all_params)
def all_vs_all_comparisons(file_chunks, 
                           job_completion_flag,
                           output_dir,
                           result_file,
                           name):
    """
    Simulate comparison of gene chunks against each  other
    Normally runs in parallel on a computational cluster       
    """

    #
    #   make output directory
    #    
    try:
        os.makedirs(output_dir)
    except OSError:
        pass
    
    open(job_completion_flag, "w")
    open(result_file, "w").write(name)


#_________________________________________________________________________________________
#
#   Step 4:
#       
#       Recombine: alignment results to make gene families
#            compare/x/*.comparison_res
#               -> multiple_alignment/x/x.gene_families
#       
#_________________________________________________________________________________________

#
#   generate_params_for_making_gene_families
#       
#           function for generating custom parameters
#
def generate_params_for_making_gene_families ():
    """
    Custom function to combining comparison files into gene families
    """
    gene_set_names = get_unknown_gene_set_names()
    for x in gene_set_names:
        results_files = glob.glob("%s/compare/%s/*.comparison_res" % (w_dir, x))
        output_dir = "%s/multiple_alignment/%s" % (w_dir, x)
        family_file = "%s/gene.families" % output_dir
        yield results_files, family_file, output_dir


@follows(all_vs_all_comparisons)
@files(generate_params_for_making_gene_families)
def combine_into_gene_familes (results_files, family_file_name, output_dir):
    """
    Simulate making gene families by concatenating comparison results :-)
    """
    #
    #   make output directory
    #    
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    family_file = open(family_file_name, "w")
    for f in results_files:
        family_file.write(open(f).read())
    sleep(1)

#_________________________________________________________________________________________
#
#   Step 5:
#       
#       split_gene_family_for_evolutionary_analysis
#           multiple_alignment/x/x.gene_families
#               -> multiple_alignment/x/NNN.aln
#               -> multiple_alignment/x/split.completed
#       
#_________________________________________________________________________________________
@follows(combine_into_gene_familes)
@files_re("%s/multiple_alignment/*/gene.families"      % w_dir,
          r"(.+)/(gene.families)",                    
          r"\1/\2",                                   
          r"\1/split.completed",
          r"\1")
def split_gene_family_for_evolutionary_analysis( family_file, 
                                                 job_completion_flag_file, split_output_dir):
    """
    Simulate splitting family of genes into "chunks" suitable for
        parallel jobs on a computational cluster       
    """

    # save number of chunks for later tasks
    number_of_output_files = int(random.uniform(20, 50))
    
    for index in range(number_of_output_files):
        open("%s/%d.aln" % (split_output_dir, index), "w").write("chunk %d" % index)
    open(job_completion_flag_file, "w")
    sleep(1)


#_________________________________________________________________________________________
#
#   Step 6:
#       
#       evolution_analysis
#           multiple_alignment/x/NNN.aln
#               -> multiple_alignment/x/NNN.evo_res
#       
#_________________________________________________________________________________________
@follows(split_gene_family_for_evolutionary_analysis)
@files_re("%s/multiple_alignment/*/*.aln"      % w_dir,
          r"(.+).aln",                    
          r"\1.evo_res")
def evolution_analysis( family_file, result_file_name):
    """
    Simulate evolutionary analysis
    """

    result_file = open(result_file_name, "w")
    result_file.write(family_file + "\n")
    sleep(1)

    
#_________________________________________________________________________________________
#
#   Step 7:
#       
#       combine_evolution_analysis
#           multiple_alignment/x/NNN.evo_res
#               -> evolutionary_analysis/x.results
#       
#_________________________________________________________________________________________
    
#
#   generate_params_for_combining_evolutionary_analyses
#       
#           function for generating custom parameters
#
def generate_params_for_combining_evolutionary_analyses ():
    """
    Custom function to combining evolutionary analyses per unknown gene set
    """
    gene_set_names = get_unknown_gene_set_names()
    for x in gene_set_names:
        results_files = glob.glob("%s/multiple_alignment/%s/*.evo_res" % (w_dir, x))
        combined_file = "%s/evolutionary_analysis/%s.results" % (w_dir, x)
        yield results_files, combined_file
        
@follows(evolution_analysis, mkdir("%s/evolutionary_analysis" % w_dir))
@files(generate_params_for_combining_evolutionary_analyses)
def combine_evolution_analysis (results_files, combined_file_name):
    """
    Simular combining evolutionary analyses
    """
    combined_file = open(combined_file_name, "w")
    for f  in results_files:
        combined_file.write(open(f).read())
    sleep(1)
    
    

#_________________________________________________________________________________________
#
#   Step 8:
#       
#       summarise_evolution_analysis
#           evolutionary_analysis/x.results
#                -> evolutionary_analysis/x.summary
#       
#_________________________________________________________________________________________
@follows(combine_evolution_analysis)
@files_re("%s/evolutionary_analysis/*.results"      % w_dir,
          r"(.+).results",                    
          r"\1.summary")
def summarise_evolution_analysis( results_file, summary_file_name):
    """
    Simulate summary of evolutionary analysis
    """
    summary_file = open(summary_file_name, "w")
    summary_file.write("summary of " + open(results_file).read())
    sleep(1)


#_________________________________________________________________________________________
#
#   Step 9:
#       
#       summarise_all
#       evolutionary_analysis/x.summary
#           -> all.total_summary
#       
#_________________________________________________________________________________________
summary_file_names = ["%s/evolutionary_analysis/%s.summary" % (w_dir, n) 
                        for n in get_unknown_gene_set_names()]
total_summary_file_name = "%s/all.total_summary" % w_dir

@follows(summarise_evolution_analysis)
@files(summary_file_names, total_summary_file_name)
def summarise_all( summary_files, total_summary_file_name):
    """
    Simulate summarize all
    """
    total_summary_file = open(total_summary_file_name, "w")
    total_summary_file.write("Over all Summary:\n")
    for f in summary_files:
        total_summary_file.write(open(f).read())
    sleep(1)



    
    
    
    
#888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888    
#
#   print pipeline or run pipeline
# 

if options.just_print:
    pipeline_printout(sys.stdout, options.target_tasks, options.forced_tasks, long_winded=True)

elif options.dependency_file:
    graph_printout (     open(options.dependency_file, "w"),
                         options.dependency_graph_format,
                         options.target_tasks, 
                         options.forced_tasks)
else:    
    pipeline_run(options.target_tasks, options.forced_tasks, multiprocess = options.jobs)
    



#! /usr/bin/env python

from __future__ import print_function
from ruffus.ruffus_utility import RUFFUS_HISTORY_FILE, CHECKSUM_FILE_TIMESTAMPS, get_default_history_file_name
from ruffus.drmaa_wrapper import run_job
from ruffus.ruffus_exceptions import RethrownJobError
import ruffus

import unittest
import os
import sys
import shutil
try:
    from StringIO import StringIO
except:
    from io import StringIO
import time
import re
import subprocess

exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
sys.path.insert(0, os.path.abspath(os.path.join(exe_path, "..", "..")))

# from ruffus.combinatorics import *


workdir = 'tmp_test_job_history_with_exceptions'
# sub-1s resolution in system?
one_second_per_job = None
throw_exception = False

@ruffus.mkdir(workdir)
@ruffus.originate([workdir + "/" + prefix + "_name.tmp1" for prefix in "abcdefghijk"])
def generate_initial_files1(on):
    with open(on, 'w') as outfile:
        pass


@ruffus.transform(generate_initial_files1,
           ruffus.suffix(".tmp1"), ".tmp2")
def test_task2(infile, outfile):
    print("%s start to run " % infile)
    run_job("./five_second.py", run_locally=True)
    print("%s wake up " % infile)
    with open(outfile, "w") as p:
        pass


@ruffus.transform(test_task2, ruffus.suffix(".tmp2"), ".tmp3")
def test_task3(infile, outfile):
    print("%s start to run " % infile)
    # subprocess.check_call("./five_second.py")
    run_job("./five_second.py", run_locally=True)
    print("%s wake up " % infile)
    with open(outfile, "w") as p:
        pass


def cleanup_tmpdir():
    os.system('rm -f %s %s' %
              (os.path.join(workdir, '*'), RUFFUS_HISTORY_FILE))


def do_main():
    print("Press Ctrl-C Now!!", file=sys.stdout)
    sys.stdout.flush()
    time.sleep(2)
    print("Start....", file=sys.stdout)
    sys.stdout.flush()
    ruffus.pipeline_run(verbose=11,
                        multiprocess=5, pipeline="main")
    print("too late!!", file=sys.stdout)
    sys.stdout.flush()
    cleanup_tmpdir()


do_main()



#!/usr/bin/env python
from __future__ import print_function
import unittest
import json
from ruffus import transform, follows, pipeline_run, Pipeline, regex, mkdir, files
import sys

"""

    test_files_decorator.py

        test @files which take single files forwards arguments as single files rather
        than as lists

"""

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# from ruffus.file_name_parameters import


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

# use simplejson in place of json for python < 2.6
# try:
# except ImportError:
#    import simplejson
#    json = simplejson

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

def check_job_io(infiles, outfiles, extra_params):
    """
    cat input files content to output files
        after writing out job parameters
    """
    # dump parameters
    params = (infiles, outfiles) + extra_params

    if isinstance(infiles, str):
        infile_names = [infiles]
    elif infiles is None:
        infile_names = []
    else:
        infile_names = infiles

    if isinstance(outfiles, str):
        outfile_names = [outfiles]
    else:
        outfile_names = outfiles

    output_text = list()
    for f in infile_names:
        with open(f) as ii:
            output_text.append(ii.read())
    output_text = "".join(sorted(output_text))
    output_text += json.dumps(infiles) + " -> " + json.dumps(outfiles) + "\n"
    for f in outfile_names:
        with open(f, "w") as oo:
            oo.write(output_text)


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

@follows(mkdir(tempdir))
#
#    task1
#
@files(None, tempdir + 'a.1')
def task1(infiles, outfiles, *extra_params):
    """
    First task
    """
    check_job_io(infiles, outfiles, extra_params)


#
#    task2
#
@transform(task1, regex(r".+"), tempdir + 'b.1')
def task2(infiles, outfiles, *extra_params):
    """
    Second task
    """
    check_job_io(infiles, outfiles, extra_params)
    assert(infiles == tempdir + "a.1")


#
#    task3
#
@files(task2, tempdir + 'c.1')
def task3(infiles, outfiles, *extra_params):
    """
    Second task
    """
    check_job_io(infiles, outfiles, extra_params)
    assert(infiles == tempdir + "b.1")

#
#    task4
#


@follows(task3)
@files([[None, tempdir + 'd.1'], [None, tempdir + 'e.1']])
def task4(infiles, outfiles, *extra_params):
    """
    Second task
    """
    check_job_io(infiles, outfiles, extra_params)

#
#    task5
#


@files(task4, tempdir + "f.1")
def task5(infiles, outfiles, *extra_params):
    """
    Second task
    """
    check_job_io(infiles, outfiles, extra_params)
    assert(infiles == [tempdir + "d.1", tempdir + "e.1"])


class Test_task(unittest.TestCase):

    def tearDown(self):
        """
        """
        import glob
        for f in glob.glob(tempdir + "*"):
            os.unlink(f)
        os.rmdir(tempdir)

    def test_task(self):
        pipeline_run(multiprocess=10, verbose=0, pipeline="main")

    def test_newstyle_task(self):
        """
        Same as above but construct a new pipeline on the fly without decorators
        """
        test_pipeline = Pipeline("test")
        test_pipeline.files(task1, None, tempdir + 'a.1')\
            .follows(mkdir(tempdir))
        test_pipeline.transform(task_func=task2,
                                input=task1,
                                filter=regex(r".+"),
                                output=tempdir + 'b.1')
        test_pipeline.files(task3, task2, tempdir + 'c.1')
        test_pipeline.files(task4, [[None, tempdir + 'd.1'], [None, tempdir + 'e.1']])\
            .follows(task3)
        test_pipeline.files(task5, task4, tempdir + "f.1")
        test_pipeline.run(multiprocess=10, verbose=0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import unittest
from ruffus.ruffus_utility import CHECKSUM_HISTORY_TIMESTAMPS
from ruffus.file_name_parameters import open_job_history
from ruffus import parallel, pipeline_run, Pipeline, task
import ruffus
import sys
import os

"""
    test_task_file_dependencies.py
"""
history_file = ':memory:'
history_file = False


# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name


class dummy_task (object):
    checksum_level = CHECKSUM_HISTORY_TIMESTAMPS

    def user_defined_work_func(self):
        pass


class Test_needs_update_check_modify_time(unittest.TestCase):

    def setUp(self):
        """
        Create a list of files separated in time so we can do dependency checking
        """
        import tempfile
        import time
        self.files = list()
        job_history = open_job_history(history_file)
        for i in range(6):
            #test_file =tempfile.NamedTemporaryFile(delete=False, prefix='testing_tmp')
            #self.files.append (test_file.name)
            # test_file.close()

            fh, temp_file_name = tempfile.mkstemp(suffix='.dot')
            self.files.append(temp_file_name)
            os.fdopen(fh, "w").close()

            # Save modify time in history file
            mtime = os.path.getmtime(temp_file_name)
            epoch_seconds = time.time()
            # Use epoch seconds unless there is a > 1 second discrepancy between system clock
            # and file system clock
            if epoch_seconds > mtime and epoch_seconds - mtime < 1.1:
                mtime = epoch_seconds
            else:
                # file system clock out of sync:
                #   Use file modify times: slow down in case of low counter resolution
                #       (e.g. old versions of NFS and windows)
                time.sleep(2)
            chksum = task.JobHistoryChecksum(
                temp_file_name, mtime, "", dummy_task())
            job_history[os.path.relpath(temp_file_name)] = chksum

    def tearDown(self):
        """
        delete files
        """
        for f in self.files:
            os.unlink(f)

    def test_up_to_date(self):
        #
        #   lists of files
        #
        self.assertTrue(not task.needs_update_check_modify_time(self.files[0:2],
                                                                self.files[2:6],
                                                                job_history=open_job_history(
                                                                    history_file),
                                                                task=dummy_task())[0])
        self.assertTrue(task.needs_update_check_modify_time(self.files[2:6],
                                                            self.files[0:2],
                                                            job_history=open_job_history(
                                                                history_file),
                                                            task=dummy_task())[0])
        #
        #   singletons and lists of files
        #
        self.assertTrue(not task.needs_update_check_modify_time(self.files[0],
                                                                self.files[2:6],
                                                                job_history=open_job_history(
                                                                    history_file),
                                                                task=dummy_task())[0])
        self.assertTrue(task.needs_update_check_modify_time(self.files[2:6],
                                                            self.files[0],
                                                            job_history=open_job_history(
                                                                history_file),
                                                            task=dummy_task())[0])

        #
        #   singletons
        #
        self.assertTrue(task.needs_update_check_modify_time(self.files[3],
                                                            self.files[0],
                                                            job_history=open_job_history(
                                                                history_file),
                                                            task=dummy_task())[0])

        # self -self = no update
        self.assertTrue(not task.needs_update_check_modify_time(self.files[0],
                                                                self.files[0],
                                                                job_history=open_job_history(
                                                                    history_file),
                                                                task=dummy_task())[0])

        #
        #   missing files means need update
        #
        self.assertTrue(task.needs_update_check_modify_time(self.files[0:2] +
                                                            ["uncreated"],
                                                            self.files[3:6],
                                                            job_history=open_job_history(
                                                                history_file),
                                                            task=dummy_task())[0])

        self.assertTrue(task.needs_update_check_modify_time(self.files[0:2],
                                                            self.files[3:6] +
                                                            ["uncreated"],
                                                            job_history=open_job_history(
                                                                history_file),
                                                            task=dummy_task())[0])

        #
        #   None means need update
        #
        self.assertTrue(task.needs_update_check_modify_time(self.files[0:2],
                                                            None,
                                                            job_history=open_job_history(
                                                                history_file),
                                                            task=dummy_task())[0])

        #
        #   None input means need update only if do not exist
        #
        self.assertTrue(not task.needs_update_check_modify_time(None,
                                                                self.files[3:6],
                                                                job_history=open_job_history(
                                                                    history_file),
                                                                task=dummy_task())[0])

        #
        #   None + missing file means need update
        #
        self.assertTrue(task.needs_update_check_modify_time(self.files[0:2] +
                                                            ["uncreated"],
                                                            None,
                                                            job_history=open_job_history(
                                                                history_file),
                                                            task=dummy_task())[0])

        self.assertTrue(task.needs_update_check_modify_time(None,
                                                            self.files[3:6] +
                                                            ["uncreated"],
                                                            job_history=open_job_history(
                                                                history_file),
                                                            task=dummy_task())[0])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import shutil
import unittest
from ruffus.ruffus_utility import RUFFUS_HISTORY_FILE, CHECKSUM_FILE_TIMESTAMPS
from ruffus import pipeline_run, pipeline_printout, transform, split, mkdir, formatter, Pipeline
import sys

"""

    test_mkdir.py
        test product, combine, permute, combine_with_replacement

"""

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0]))


# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#   imports
#
# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

try:
    from StringIO import StringIO
except:
    from io import StringIO


# sub-1s resolution in system?
# ___________________________________________________________________________
#
#   generate_initial_files1
# ___________________________________________________________________________
@split(1, [tempdir + "/" + prefix + "_name.tmp1" for prefix in "abcd"])
def generate_initial_files1(in_name, out_names):
    for on in out_names:
        with open(on, 'w') as outfile:
            pass

# ___________________________________________________________________________
#
#   check_product_task
# ___________________________________________________________________________


@mkdir(tempdir + "/test1")
@mkdir(tempdir + "/test2")
@mkdir(generate_initial_files1, formatter(),
       ["{path[0]}/{basename[0]}.dir", 3, "{path[0]}/{basename[0]}.dir2"])
@transform(generate_initial_files1,
           formatter(),
           "{path[0]}/{basename[0]}.dir/{basename[0]}.tmp2")
def check_transform(infiles, outfile):
    with open(outfile, "w") as p:
        pass


@mkdir(tempdir + "/test3")
@mkdir(generate_initial_files1, formatter(),
       "{path[0]}/{basename[0]}.dir2")
def check_transform2():
    print("    Loose cannon!", file=sys.stderr)


def cleanup_tmpdir():
    os.system('rm -f %s %s' %
              (os.path.join(tempdir, '*'), RUFFUS_HISTORY_FILE))


class Testmkdir(unittest.TestCase):
    def setUp(self):
        try:
            os.mkdir(tempdir)
        except OSError:
            pass

    # ___________________________________________________________________________
    #
    #   test mkdir() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_mkdir_printout(self):
        """Input file exists, output doesn't exist"""
        cleanup_tmpdir()

        s = StringIO()
        pipeline_printout(s, [check_transform, check_transform2],
                          verbose=5, wrap_width=10000, pipeline="main")
        # self.assertIn('Job needs update: Missing files '
        #              '[tmp_test_mkdir/a_name.tmp1, '
        #              'tmp_test_mkdir/e_name.tmp1, '
        #              'tmp_test_mkdir/h_name.tmp1, '
        #              'tmp_test_mkdir/a_name.e_name.h_name.tmp2]', s.getvalue())

    def test_mkdir_run(self):
        """Run mkdir"""
        # output is up to date, but function body changed (e.g., source different)
        cleanup_tmpdir()
        pipeline_run([check_transform, check_transform2],
                     verbose=0, multiprocess=2, pipeline="main")

    def test_newstyle_mkdir_run(self):
        test_pipeline = Pipeline("test")

        test_pipeline.split(task_func=generate_initial_files1,
                            input=1,
                            output=[tempdir + "/" + prefix + "_name.tmp1" for prefix in "abcd"])

        test_pipeline.transform(task_func=check_transform,
                                input=generate_initial_files1,
                                filter=formatter(),
                                output="{path[0]}/{basename[0]}.dir/{basename[0]}.tmp2")\
            .mkdir(tempdir + "/test1")\
            .mkdir(tempdir + "/test2")\
            .mkdir(generate_initial_files1, formatter(),
                   ["{path[0]}/{basename[0]}.dir", 3, "{path[0]}/{basename[0]}.dir2"])

        test_pipeline.mkdir(check_transform2, tempdir + "/test3")\
            .mkdir(generate_initial_files1, formatter(),
                   "{path[0]}/{basename[0]}.dir2")
        cleanup_tmpdir()
        pipeline_run([check_transform, check_transform2],
                     verbose=0, multiprocess=2, pipeline="main")

    # ___________________________________________________________________________
    #
    #   cleanup
    # ___________________________________________________________________________

    def tearDown(self):
        shutil.rmtree(tempdir)


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import sys, os

# add self to search path for testing
if __name__ == '__main__':
    exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
    module_name = os.path.split(sys.argv[0])[1]
    module_name = os.path.splitext(module_name)[0];
else:
    module_name = __name__

# Use import path from <<../python_modules>>
if __name__ == '__main__':
    sys.path.insert(0, os.path.abspath(os.path.join(exe_path,"../..")))

from ruffus import *
import json

import time
def task_helper(infile, outfile):
    """
    cat input file content to output file
        after writing out job parameters
    """
    if infile:
        output_text = "".join(sorted(open(infile).readlines()))
    else:
        output_text = "None"
    output_text += json.dumps(infile) + " -> " + json.dumps(outfile) + "\n"
    open(outfile, "w").write(output_text)



#
#    task1
#
@files(None, 'a.1')
def task1(infile, outfile):
    """
    First task
    """
    task_helper(infile, outfile)



#
#    task2
#
@transform(task1, regex(r'.1'), '.2')
def task2(infile, outfile):
    """
    Second task
    """
    task_helper(infile, outfile)



#
#    task3
#
@transform(task2, regex(r'.2'), '.3')
def task3(infile, outfile):
    """
    Third task
    """
    task_helper(infile, outfile)



#
#    task4
#
@transform(task3, regex(r'.3'), '.4')
def task4(infile, outfile):
    """
    Fourth task
    """
    task_helper(infile, outfile)

for f in range(4):
    fn = "a.%d" % (f + 1)
    if os.path.exists(fn):
        os.unlink(fn)
pipeline_printout_graph ("manual_dependencies_flowchart1.png", "png", [task4], dpi = 72, size = (2,2))
pipeline_run([task4])
pipeline_printout_graph ("manual_dependencies_flowchart2.png", "png", [task4], dpi = 72, size = (2,2))
open("a.1", "w")
open("a.3", "w")
pipeline_printout_graph ("manual_dependencies_flowchart3.png", "png", [task4], dpi = 72, size = (2,2))
pipeline_printout_graph ("manual_dependencies_flowchart4.png", "png", [task4], [task1], dpi = 72, size = (2,2))
for f in range(4):
    fn = "a.%d" % (f + 1)
    if os.path.exists(fn):
        os.unlink(fn)




#!/usr/bin/env python
"""

    intermediate_example.py
    
        This script takes N pairs of input file pairs
                                (with the suffices .gene and .gwas)
                          and runs them against M sets of simulation data
                                (with the suffix .simulation)
        A summary per input file pair is then produced
                    
        
    In pseudo-code:
        
        STEP_1:
        
        for n_file in NNN_pairs_of_input_files:
            for m_file in MMM_simulation_data:

                [n_file.gene,
                 n_file.gwas,
                 m_file.simulation] -> n_file.m_file.simulation_res
                 

        STEP_2:
        
        for n_file in NNN_pairs_of_input_files:
            
            n_file.*.simulation_res -> n_file.mean
            
        

"""

import os, sys
exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
from ruffus import *
from time import sleep
import random
from itertools import izip


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options        


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

from optparse import OptionParser
parser = OptionParser(version="%prog 1.0")
parser.add_option("-t", "--target_tasks", dest="target_tasks",
                  action="append",
                  default = ["statistical_summary"],
                  metavar="JOBNAME", 
                  type="string",
                  help="Target task(s) of pipeline.")
parser.add_option("-f", "--forced_tasks", dest="forced_tasks",
                  action="append",
                  default = list(),
                  metavar="JOBNAME", 
                  type="string",
                  help="Pipeline task(s) which will be included even if they are up to date.")
parser.add_option("-j", "--jobs", dest="jobs",
                  default=5,
                  metavar="jobs", 
                  type="int",
                  help="Specifies the number of jobs (commands) to run simultaneously.")

parser.add_option("-g", "--gene_data_dir", dest="gene_data_dir",
                  default="%s/data_for_intermediate_example/genes" % exe_path,
                  metavar="PATH", 
                  type="string",
                  help="Directory with gene data [*.genes / *.gwas].")
parser.add_option("-s", "--simulation_data_dir", dest="simulation_data_dir",
                  default="%s/data_for_intermediate_example/simulation" % exe_path,
                  metavar="PATH", 
                  type="string",
                  help="Directory with simulation data [*.simulation].")
parser.add_option("-w", "--working_dir", dest="working_dir",
                  default="/working_dir",
                  metavar="PATH", 
                  type="string",
                  help="Working directory.")


parser.add_option("-v", "--verbose", dest = "verbose",
                  action="store_true", default=False,
                  help="Do not echo to shell but only print to log.")
parser.add_option("-D", "--dependency", dest="dependency_file",
                  metavar="FILE", 
                  type="string",
                  help="Print a dependency graph of the pipeline that would be executed "
                        "to FILE, but do not execute it.")
parser.add_option("-F", "--dependency_graph_format", dest="dependency_graph_format",
                  metavar="FORMAT", 
                  type="string",
                  default = 'svg',
                  help="format of dependency graph file. Can be 'ps' (PostScript), "+
                  "'svg' 'svgz' (Structured Vector Graphics), " +
                  "'png' 'gif' (bitmap  graphics) etc ")
parser.add_option("-n", "--just_print", dest="just_print",
                    action="store_true", default=False,
                    help="Print a description of the jobs that would be executed, "
                        "but do not execute them.")





#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports        


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

import StringIO
import re
import operator
import sys
from collections import defaultdict
import glob
#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Functions


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


#_________________________________________________________________________________________
#
#   get gene gwas file pairs
#
#_________________________________________________________________________________________
def get_gene_gwas_file_pairs(  ):
    """
    Helper function to get all *.gene, *.gwas from the direction specified
        in --gene_data_dir
        
    Returns 
        file pairs with both .gene and .gwas extensions, 
        corresponding roots (no extension) of each file
    """
    

    gene_files = glob.glob(os.path.join(options.gene_data_dir, "*.gene"))
    gwas_files = glob.glob(os.path.join(options.gene_data_dir, "*.gwas"))
    
    common_roots = set(map(lambda x: os.path.splitext(os.path.split(x)[1])[0], gene_files))
    common_roots &=set(map(lambda x: os.path.splitext(os.path.split(x)[1])[0], gwas_files))
    common_roots = list(common_roots)
        
    p = os.path; g_dir = options.gene_data_dir
    
    file_pairs = [[p.join(g_dir, x + ".gene"), p.join(g_dir, x + ".gwas")] for x in common_roots]
    
    return file_pairs, common_roots

#_________________________________________________________________________________________
#
#   get simulation files
#
#_________________________________________________________________________________________
def get_simulation_files(  ):
    """
    Helper function to get all *.simulation from the direction specified
        in --simulation_data_dir
        Returns 
            file with .simulation extensions, 
            corresponding roots (no extension) of each file
    """
    simulation_files = glob.glob(os.path.join(options.simulation_data_dir, "*.simulation"))
    simulation_roots =map(lambda x: os.path.splitext(os.path.split(x)[1])[0], simulation_files)
    return simulation_files, simulation_roots



#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888





# get help string
f =StringIO.StringIO()
parser.print_help(f)
helpstr = f.getvalue()
(options, remaining_args) = parser.parse_args()


working_dir = options.working_dir




#_________________________________________________________________________________________
#
#   Step 1:
#       
#        for n_file in NNN_pairs_of_input_files:
#            for m_file in MMM_simulation_data:
#
#                [n_file.gene,
#                 n_file.gwas,
#                 m_file.simulation] -> working_dir/n_file.m_file.simulation_res
#
#_________________________________________________________________________________________
def generate_simulation_params ():
    """
    Custom function to generate 
    file names for gene/gwas simulation study
    """

    simulation_files, simulation_file_roots    = get_simulation_files()
    gene_gwas_file_pairs, gene_gwas_file_roots =  get_gene_gwas_file_pairs()
    
    for sim_file, sim_file_root in izip(simulation_files, simulation_file_roots):
        for (gene, gwas), gene_file_root in izip(gene_gwas_file_pairs, gene_gwas_file_roots):
            
            result_file = "%s.%s.simulation_res" % (gene_file_root, sim_file_root)
            result_file_path = os.path.join(working_dir, "simulation_results", result_file)
            
            yield [gene, gwas, sim_file], result_file_path, gene_file_root, sim_file_root, result_file

# 
# mkdir: makes sure output directories exist before task    
#
@follows(mkdir(options.working_dir, os.path.join(working_dir, "simulation_results")))
@files(generate_simulation_params)
def gwas_simulation(input_files, result_file_path, gene_file_root, sim_file_root, result_file):
    """
    Dummy calculation of gene gwas vs simulation data
    Normally runs in parallel on a computational cluster       
    """
    (gene_file,
    gwas_file,
    simulation_data_file) = input_files

    simulation_res_file = open(result_file_path, "w")
    simulation_res_file.write("%s + %s -> %s\n" % (gene_file_root, sim_file_root, result_file))


#_________________________________________________________________________________________
#
#   Step 2:
#       
#       Statistical summary per gene/gwas file pair
# 
#        for n_file in NNN_pairs_of_input_files:
#            working_dir/simulation_results/n.*.simulation_res
#               -> working_dir/n.mean
#       
#_________________________________________________________________________________________
def generate_statistical_summary_params():
    """
    Custom function to summarising simulation results files per gene / gwas file pair
    """
    gene_gwas_file_pairs, gene_gwas_file_roots =  get_gene_gwas_file_pairs()

    for (gene, gwas), gene_file_root in izip(gene_gwas_file_pairs, gene_gwas_file_roots):
            result_glob_spec = "%s.*.simulation_res" % (gene_file_root)
            result_files     = glob.glob(os.path.join(working_dir, "simulation_results", result_glob_spec))
            summary_file     = os.path.join(working_dir, gene_file_root + ".mean")
                
            yield result_files, summary_file
                


@follows(gwas_simulation)
@files(generate_statistical_summary_params)
@posttask(lambda : sys.stdout.write("\nAll finished: hooray!!!\n"))
def statistical_summary (result_files, summary_file):
    """
    Simulate statistical summary
    """

    summary_file = open(summary_file, "w")
    for f in result_files:
        summary_file.write(open(f).read())
    sleep(1)
    





    
    
    
    
#888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888    
#
#   print pipeline or run pipeline
# 

if options.just_print:
    pipeline_printout(sys.stdout, options.target_tasks, options.forced_tasks, long_winded=True)

elif options.dependency_file:
    graph_printout (     open(options.dependency_file, "w"),
                         options.dependency_graph_format,
                         options.target_tasks, 
                         options.forced_tasks)
else:    
    pipeline_run(options.target_tasks, options.forced_tasks, multiprocess = options.jobs)
    



#!/usr/bin/env python
from __future__ import print_function
import unittest
import json
from ruffus import follows, posttask, files, suffix, mkdir, Pipeline, touch_file, originate, pipeline_run
import sys

"""

    test_follows_mkdir.py

        test make directory dependencies

        use :
            -j N / --jobs N       to speify multitasking
            -v                    to see the jobs in action
            -n / --just_print     to see what jobs would run

"""


import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


def touch(filename):
    with open(filename, "w"):
        pass


if sys.hexversion >= 0x03000000:
    unicode = str

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
directories = [os.path.abspath(unicode(tempdir + "a")), unicode(tempdir + "b")]


@follows(mkdir(directories), mkdir(unicode(tempdir + "c")), mkdir(unicode(tempdir + "d"), unicode(tempdir + "e")), mkdir(unicode(tempdir + "e")))
@posttask(touch_file(unicode(tempdir + "f")))
def task_which_makes_directories():
    pass


@originate([tempdir + "g", tempdir + "h"])
def task_which_makes_files(o):
    touch(o)


class Test_task_mkdir(unittest.TestCase):

    def setUp(self):
        """
        """
        os.makedirs(tempdir)
        pass

    def tearDown(self):
        """
        delete directories
        """
        for d in 'abcde':
            fullpath = os.path.join(os.path.dirname(__file__), tempdir, d)
            os.rmdir(fullpath)
        for d in 'fgh':
            fullpath = os.path.join(os.path.dirname(__file__), tempdir, d)
            os.unlink(fullpath)

        os.rmdir(tempdir)

    def test_mkdir(self):
        pipeline_run(multiprocess=10, verbose=0, pipeline="main")

        for d in 'abcdefgh':
            fullpath = os.path.join(os.path.dirname(__file__), tempdir, d)
            self.assertTrue(os.path.exists(fullpath))

    def test_newstyle_mkdir(self):
        test_pipeline = Pipeline("test")

        test_pipeline.follows(task_which_makes_directories,
                              mkdir(directories),
                              mkdir(unicode(tempdir + "c")),
                              mkdir(unicode(tempdir + "d"),
                                    unicode(tempdir + "e")),
                              mkdir(unicode(tempdir + "e")))\
            .posttask(touch_file(unicode(tempdir + "f")))

        test_pipeline.originate(task_which_makes_files, [
                                tempdir + "g", tempdir + "h"])
        test_pipeline.run(multiprocess=10, verbose=0)

        for d in 'abcdefgh':
            fullpath = os.path.join(os.path.dirname(__file__), tempdir, d)
            self.assertTrue(os.path.exists(fullpath))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import json
import shutil
import unittest
from ruffus import transform, subdivide, merge, suffix, mkdir, pipeline_run, Pipeline, originate
import ruffus
import sys

"""

    test_suffix_output_dir.py

"""

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0]))
data_dir = tempdir + "/data"
work_dir = tempdir + "/work"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

def helper(infiles, outfiles):
    """
    Helper function: Writes input output file names and input contents to outfile
    """
    # None = []
    if outfiles is None:
        outfiles = []
    # str = [str]
    if not isinstance(outfiles, list):
        outfiles = [outfiles]

    for outfile in outfiles:
        with open(outfile, "w") as oo:
            # save file name strings before we turn infiles into a list
            fn_str = "%s -> %s" % (infiles, outfile)

            # None = []
            if infiles is None:
                infiles = []
            # str = [str]
            if not isinstance(infiles, list):
                infiles = [infiles]

            # write header
            # if len(infiles):
            #    for infile in infiles:
            #        with open(infile) as ii:
            #            for line in ii:
            #                line = line.rstrip()
            #                if line[-1:] == ":":
            #                    oo.write(outfile + ":" + line + "\n")
            # else:
            #    oo.write(outfile + ":\n")

            max_white_space = -2
            # write content of infiles indented
            for infile in infiles:
                with open(infile) as ii:
                    for line in ii:
                        line = line.rstrip()
                        if line[-1:] != ":":
                            oo.write(line + "\n")
                        max_white_space = max(
                            [max_white_space, len(line) - len(line.lstrip())])

            # add extra spaces before filenames
            oo.write(" " * (max_white_space + 2) + fn_str + "\n")


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#
#    task1
#
@mkdir(data_dir, work_dir)
@originate([os.path.join(data_dir, "%s.1" % aa) for aa in "abcd"])
def task1(outfile):
    """
    originate
    """
    # N.B. originate works with an extra parameter
    helper(None, outfile)


#
#    task2
#
@mkdir(task1,
       suffix(".1"),
       ".dir",
       output_dir=work_dir)
@transform(task1, suffix(".1"), [".1", ".bak"], "extra.tst", 4, r"orig_dir=\1", output_dir=work_dir)
def task2(infile, outfile, extra_str, extra_num, extra_dir):
    """
    transform switch directory
    """
    if (extra_str, extra_num) != ("extra.tst", 4) or \
       extra_dir[:len("orig_dir=" + work_dir)] != "orig_dir=" + work_dir:
        raise Exception("transform with output_dir has changed extras: %s, %s %s"
                        % (extra_str, extra_num, extra_dir))
    helper(infile, outfile)


#
#    task3
#
@subdivide(task2, suffix(".1"), r"\1.*.2", [r"\1.a.2", r"\1.b.2"], output_dir=data_dir)
def task3(infile, ignore_outfiles, outfiles):
    """
    subdivide
    """
    helper(infile, outfiles)


#
#    task4
#
@transform(task3, suffix(".2"), ".3", output_dir=work_dir)
def task4(infile, outfile):
    """
    transform
    """
    helper(infile, outfile)

#
#    task4
#


@merge(task4, os.path.join(data_dir, "summary.5"))
def task5(infiles, outfile):
    """
    merge
    """
    helper(infiles, outfile)


expected_active_text = \
    """None -> {tempdir}/data/a.1
  {tempdir}/data/a.1 -> {tempdir}/work/a.1
None -> {tempdir}/data/a.1
  ['{tempdir}/data/a.1'] -> {tempdir}/work/a.bak
    ['{tempdir}/work/a.1', '{tempdir}/work/a.bak'] -> {tempdir}/data/a.a.2
      {tempdir}/data/a.a.2 -> {tempdir}/work/a.a.3
None -> {tempdir}/data/a.1
  {tempdir}/data/a.1 -> {tempdir}/work/a.1
None -> {tempdir}/data/a.1
  ['{tempdir}/data/a.1'] -> {tempdir}/work/a.bak
    ['{tempdir}/work/a.1', '{tempdir}/work/a.bak'] -> {tempdir}/data/a.b.2
      {tempdir}/data/a.b.2 -> {tempdir}/work/a.b.3
None -> {tempdir}/data/b.1
  {tempdir}/data/b.1 -> {tempdir}/work/b.1
None -> {tempdir}/data/b.1
  ['{tempdir}/data/b.1'] -> {tempdir}/work/b.bak
    ['{tempdir}/work/b.1', '{tempdir}/work/b.bak'] -> {tempdir}/data/b.a.2
      {tempdir}/data/b.a.2 -> {tempdir}/work/b.a.3
None -> {tempdir}/data/b.1
  {tempdir}/data/b.1 -> {tempdir}/work/b.1
None -> {tempdir}/data/b.1
  ['{tempdir}/data/b.1'] -> {tempdir}/work/b.bak
    ['{tempdir}/work/b.1', '{tempdir}/work/b.bak'] -> {tempdir}/data/b.b.2
      {tempdir}/data/b.b.2 -> {tempdir}/work/b.b.3
None -> {tempdir}/data/c.1
  {tempdir}/data/c.1 -> {tempdir}/work/c.1
None -> {tempdir}/data/c.1
  ['{tempdir}/data/c.1'] -> {tempdir}/work/c.bak
    ['{tempdir}/work/c.1', '{tempdir}/work/c.bak'] -> {tempdir}/data/c.a.2
      {tempdir}/data/c.a.2 -> {tempdir}/work/c.a.3
None -> {tempdir}/data/c.1
  {tempdir}/data/c.1 -> {tempdir}/work/c.1
None -> {tempdir}/data/c.1
  ['{tempdir}/data/c.1'] -> {tempdir}/work/c.bak
    ['{tempdir}/work/c.1', '{tempdir}/work/c.bak'] -> {tempdir}/data/c.b.2
      {tempdir}/data/c.b.2 -> {tempdir}/work/c.b.3
None -> {tempdir}/data/d.1
  {tempdir}/data/d.1 -> {tempdir}/work/d.1
None -> {tempdir}/data/d.1
  ['{tempdir}/data/d.1'] -> {tempdir}/work/d.bak
    ['{tempdir}/work/d.1', '{tempdir}/work/d.bak'] -> {tempdir}/data/d.a.2
      {tempdir}/data/d.a.2 -> {tempdir}/work/d.a.3
None -> {tempdir}/data/d.1
  {tempdir}/data/d.1 -> {tempdir}/work/d.1
None -> {tempdir}/data/d.1
  ['{tempdir}/data/d.1'] -> {tempdir}/work/d.bak
    ['{tempdir}/work/d.1', '{tempdir}/work/d.bak'] -> {tempdir}/data/d.b.2
      {tempdir}/data/d.b.2 -> {tempdir}/work/d.b.3
        ['{tempdir}/work/a.a.3', '{tempdir}/work/a.b.3', '{tempdir}/work/b.a.3', '{tempdir}/work/b.b.3', '{tempdir}/work/c.a.3', '{tempdir}/work/c.b.3', '{tempdir}/work/d.a.3', '{tempdir}/work/d.b.3'] -> {tempdir}/data/summary.5
""".format(tempdir=tempdir)


class Test_ruffus(unittest.TestCase):
    def setUp(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass
        for dd in tempdir, work_dir, data_dir:
            try:
                os.makedirs(dd)
            except:
                pass

    def tearDown(self):
        try:
            shutil.rmtree(tempdir)
        except:
            sys.stderr.write("Can't remove %s" % tempdir)
            pass

    def test_ruffus(self):
        pipeline_run(multiprocess=50, verbose=0, pipeline="main")

        with open(os.path.join(data_dir, "summary.5")) as ii:
            active_text = ii.read()
        if active_text != expected_active_text:
            raise Exception("Error:\n\tExpected\n%s\nInstead\n%s\n" %
                            (expected_active_text, active_text))

    def test_newstyle_ruffus(self):
        # alternative syntax
        test_pipeline = Pipeline("test")

        test_pipeline.mkdir(data_dir, work_dir)
        test_pipeline.originate(task_func=task1,
                                output=[os.path.join(data_dir, "%s.1" % aa) for aa in "abcd"])

        test_pipeline.mkdir(filter=suffix(".1"),
                            output=".dir",
                            output_dir=work_dir)

        test_pipeline.transform(task_func=task2,
                                input=task1,
                                filter=suffix(".1"),
                                output=[".1", ".bak"],
                                extras=["extra.tst", 4, r"orig_dir=\1"],
                                output_dir=work_dir)

        test_pipeline.subdivide(task3, task2, suffix(
            ".1"), r"\1.*.2", [r"\1.a.2", r"\1.b.2"], output_dir=data_dir)
        test_pipeline.transform(task4, task3, suffix(
            ".2"), ".3", output_dir=work_dir)
        test_pipeline.merge(task5, task4, os.path.join(data_dir, "summary.5"))
        test_pipeline.run(multiprocess=50, verbose=0)

        with open(os.path.join(data_dir, "summary.5")) as ii:
            active_text = ii.read()
        if active_text != expected_active_text:
            raise Exception("Error:\n\tExpected\n%s\nInstead\n%s\n" %
                            (expected_active_text, active_text))


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
################################################################################
#
#
#   drmaa_wrapper.py
#
#   Copyright (C) 2013 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#
#   Portions of code from adapted from:
#
#       http://stackoverflow.com/questions/375427/non-blocking-read-on-a-subprocess-pipe-in-python
#       Courtesy of J.F. Sebastian
#       Use is licensed under the "Creative Commons Attribution Share Alike license"
#       See http://stackexchange.com/legal
#
#################################################################################
"""

********************************************
:mod:`ruffus.cmdline` -- Overview
********************************************

.. moduleauthor:: Leo Goodstadt <ruffus@llew.org.uk>


    #
    #   Using drmaa
    #
    from ruffus import *
    import drmaa_wrapper

"""

import sys
import os
import stat
import tempfile
import datetime
import subprocess
import time
import threading
from .task import lookup_pipeline
from .ruffus_exceptions import JobSignalledBreak, JobSignalledSuspend, JobSignalledResume, \
    JobFailed

try:
    import gevent
    HAVE_GEVENT = True
except ImportError:
    HAVE_GEVENT = False


try:
    import drmaa
    HAVE_DRMAA = True
except ImportError:
    HAVE_DRMAA = False

try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty  # python 3.x

ON_POSIX = 'posix' in sys.builtin_module_names

# Timeouts for gevent loop
GEVENT_TIMEOUT_STARTUP = 5
GEVENT_TIMEOUT_WAIT = 1


if sys.hexversion >= 0x03000000:
    # everything is unicode in python3
    path_str_type = str
else:
    path_str_type = basestring


class error_drmaa_job(Exception):
    """
    All exceptions throw in this module
    """

    def __init__(self, *errmsg):
        Exception.__init__(self, *errmsg)


def read_stdout_stderr_from_files(stdout_path, stderr_path, logger=None, cmd_str="", tries=5):
    """
    Reads the contents of two specified paths and returns the strings

    Thanks to paranoia approach contributed by Andreas Heger:

        Retry just in case file system hasn't committed.

        Logs error if files are missing: No big deal?

        Cleans up files afterwards

        Returns tuple of stdout and stderr.

    """
    # delay up to 10 seconds until files are ready
    for xxx in range(tries):
        if os.path.exists(stdout_path) and os.path.exists(stderr_path):
            break
        time.sleep(2)

    try:
        stdout = open(stdout_path, "r").readlines()
    except IOError:
        exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
        msg = str(exceptionValue)
        if logger:
            logger.warning("could not open stdout: %s for \n%s" %
                           (msg, cmd_str))
        stdout = []

    try:
        stderr = open(stderr_path, "r").readlines()
    except IOError:
        exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
        msg = str(exceptionValue)
        if logger:
            logger.warning("could not open stderr: %s for \n%s" %
                           (msg, cmd_str))
        stderr = []

    # cleanup ignoring errors
    try:
        os.unlink(stdout_path)
        os.unlink(stderr_path)
    except OSError:
        pass

    return stdout, stderr


def setup_drmaa_job(drmaa_session, job_name, job_environment, working_directory, job_other_options):

    job_template = drmaa_session.createJobTemplate()

    if not working_directory:
        working_directory = os.getcwd()

    job_template.workingDirectory = working_directory

    if job_environment:
        # dictionary e.g. { 'BASH_ENV' : '~/.bashrc' }
        job_template.jobEnvironment = job_environment

    job_template.args = []

    # nameless jobs sometimes break drmaa implementations...
    if not job_name:
        job_name = "ruffus_job_" + "_".join(map(str, datetime.datetime.now().timetuple()[0:6]))

    job_template.jobName = job_name

    # optional job parameters
    if job_other_options is not None:
        job_template.nativeSpecification = job_other_options

    # separate stdout and stderr
    job_template.joinFiles = False

    return job_template


def write_job_script_to_temp_file(cmd_str, job_script_directory, job_name, job_other_options,
                                  job_environment, working_directory):
    '''
        returns (job_script_path, stdout_path, stderr_path)

    '''
    time_stmp_str = "_".join(
        map(str, datetime.datetime.now().timetuple()[0:6]))
    if job_name is None:
        job_name = 'drmaa_script_'
    # create script directory if necessary
    # Ignore errors rather than test for existence to avoid race conditions
    try:
        os.makedirs(job_script_directory)
    except:
        pass
    tmpfile = tempfile.NamedTemporaryFile(
        mode='w', prefix=job_name + "_" + time_stmp_str + "__", dir=job_script_directory,  delete=False)

    #
    #   hopefully #!/bin/sh is universally portable among unix-like operating systems
    #
    tmpfile.write("#!/bin/sh\n")
    #
    # log parameters as suggested by Bernie Pope
    #
    for title, parameter in (("job_name",             job_name,),
                             ("job_other_options",    job_other_options,),
                             ("job_environment",      job_environment,),
                             ("working_directory",     working_directory),):
        if parameter:
            tmpfile.write("#%s=%s\n" % (title, parameter))

    tmpfile.write(cmd_str + "\n")
    tmpfile.close()

    job_script_path = os.path.abspath(tmpfile.name)
    stdout_path = job_script_path + ".stdout"
    stderr_path = job_script_path + ".stderr"

    os.chmod(job_script_path, stat.S_IRWXG | stat.S_IRWXU)

    return (job_script_path, stdout_path, stderr_path)


def submit_drmaa_job(cmd_str, drmaa_session, job_template, logger, pipeline):

    jobid = drmaa_session.runJob(job_template)
    if logger:
        logger.debug("job has been submitted with jobid {}".format(jobid))

    try:
        gevent.sleep(GEVENT_TIMEOUT_STARTUP)
    except JobSignalledBreak:
        if logger:
            logger.debug("job with jobid {} will be terminated".format(jobid))
            drmaa_session.control(jobid, drmaa.JobControlAction.TERMINATE)
        raise

    job_info = None
    is_suspended = False
    while True:
        try:
            status = drmaa_session.jobStatus(jobid)
        except Exception:
            exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
            msg = str(exceptionValue)
            # ignore message 24 in PBS code 24: drmaa: Job finished
            # but resource usage information and/or termination status
            # could not be provided.":
            if not msg.startswith("code 24"):
                raise
            if logger:
                logger.debug(msg)

        if status == drmaa.JobState.DONE:
            break
        elif status == drmaa.JobState.FAILED:
            raise JobFailed("job {} failed".format(jobid))

        try:
            gevent.sleep(GEVENT_TIMEOUT_WAIT)
        except JobSignalledBreak:
            drmaa_session.control(jobid, drmaa.JobControlAction.TERMINATE)
            raise

        if not is_suspended and pipeline.is_job_suspended():
            if logger:
                logger.debug("job with jobid {} will be suspended".format(jobid))
            if status == drmaa.JobState.RUNNING:
                drmaa_session.control(jobid, drmaa.JobControlAction.SUSPEND)
            elif status == drmaa.JobState.QUEUED_ACTIVE:
                drmaa_session.control(jobid, drmaa.JobControlAction.HOLD)
            is_suspended = True
        elif is_suspended and not pipeline.is_job_suspended():
            if logger:
                logger.debug("job with jobid {} will be resumed".format(jobid))
            if status == drmaa.JobState.USER_SUSPENDED:
                drmaa_session.control(jobid, drmaa.JobControlAction.RESUME)
            elif status == drmaa.JobState.USER_ON_HOLD:
                drmaa_session.control(jobid, drmaa.JobControlAction.RELEASE)
            is_suspended = False

    return jobid, job_info


def run_job_using_drmaa(cmd_str, job_name=None, job_other_options=None,
                        job_script_directory=None, job_environment=None,
                        working_directory=None, retain_job_scripts=False, logger=None,
                        drmaa_session=None, verbose=0, resubmit=0,
                        pipeline=None):
    """
    Runs specified command remotely using drmaa,
    either with the specified session, or the module shared drmaa session
    """
    # used specified session else module session
    if drmaa_session is None:
        raise error_drmaa_job("Please specify a drmaa_session in run_job()")

    # make job template
    job_template = setup_drmaa_job(
        drmaa_session, job_name, job_environment, working_directory, job_other_options)

    # make job script
    if not job_script_directory:
        job_script_directory = os.getcwd()
    job_script_path, stdout_path, stderr_path = write_job_script_to_temp_file(
        cmd_str, job_script_directory, job_name, job_other_options, job_environment, working_directory)
    job_template.remoteCommand = job_script_path

    # drmaa paths specified as [hostname]:file_path.
    # See http://www.ogf.org/Public_Comment_Docs/Documents/2007-12/ggf-drmaa-idl-binding-v1%2000%20RC7.pdf
    job_template.outputPath = ":" + stdout_path
    job_template.errorPath = ":" + stderr_path

    # Run job and wait
    if resubmit:
        exitStatus = 1
        jobCount = 0
        while (exitStatus and jobCount < resubmit):
            try:
                jobid, job_info = submit_drmaa_job(
                    cmd_str, drmaa_session, job_template, logger, pipeline)
                if job_info:
                    exitStatus = job_info.exitStatus
                    if exitStatus:
                        if logger:
                            logger.debug("job %s exited with %d" %
                                         (str(jobid),  exitStatus))
                    else:
                        if logger:
                            logger.debug("job %s exited normally" % str(jobid))
            except Exception as err:
                if logger:
                    logger.debug(err)
                exitStatus = 1

            jobCount += 1
            if logger and exitStatus:
                logger.debug(
                    "Resubmitting job, resubmission count is %d" % jobCount)
    else:
        jobid, job_info = submit_drmaa_job(
            cmd_str, drmaa_session, job_template, logger, pipeline)

    #
    #   Read output
    #
    stdout, stderr = read_stdout_stderr_from_files(
        stdout_path, stderr_path, logger, cmd_str)

    job_info_str = ("The original command was: >> %s <<\n"
                    "The jobid was: %s\n"
                    "The job script name was: %s\n" %
                    (cmd_str,
                     jobid,
                     job_script_path))

    def stderr_stdout_to_str(stderr, stdout):
        """
        Concatenate stdout and stderr to string
        """
        result = ""
        if stderr:
            result += "The stderr was: \n%s\n\n" % ("".join(stderr))
        if stdout:
            result += "The stdout was: \n%s\n\n" % ("".join(stdout))
        return result

    #   Throw if failed
    if job_info:
        job_info_str += "Resources used: %s " % (job_info.resourceUsage)
        if job_info.wasAborted:
            raise error_drmaa_job("The drmaa command was never ran but used %s:\n%s"
                                  % (job_info.exitStatus, job_info_str + stderr_stdout_to_str(stderr, stdout)))
        elif job_info.hasSignal:
            raise error_drmaa_job("The drmaa command was terminated by signal %i:\n%s"
                                  % (job_info.exitStatus, job_info_str + stderr_stdout_to_str(stderr, stdout)))
        elif job_info.hasExited:
            if job_info.exitStatus:
                raise error_drmaa_job("The drmaa command was terminated by signal %i:\n%s"
                                      % (job_info.exitStatus, job_info_str + stderr_stdout_to_str(stderr, stdout)))
            #   Decorate normal exit with some resource usage information
            elif verbose:
                def nice_mem_str(num):
                    """
                    Format memory sizes
                    http://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
                    """
                    num = float(num)
                    for x in ['bytes', 'KB', 'MB', 'GB']:
                        if num < 1024.0:
                            return "%3.1f%s" % (num, x)
                        num /= 1024.0
                    return "%3.1f%s" % (num, 'TB')

                try:
                    resource_usage_str = []
                    if 'maxvmem' in job_info.resourceUsage:
                        if 'mem' in job_info.resourceUsage:
                            resource_usage_str.append("Mem=%s(%s)" % (nice_mem_str(
                                job_info.resourceUsage['maxvmem']), job_info.resourceUsage['mem']))
                        else:
                            resource_usage_str.append("Mem=%s" % nice_mem_str(
                                job_info.resourceUsage['maxvmem']))
                    if 'ru_wallclock' in job_info.resourceUsage:
                        resource_usage_str.append("CPU wallclock= %.2gs" % float(
                            job_info.resourceUsage['ru_wallclock']))
                    if len(resource_usage_str):
                        logger.info("Drmaa command used %s in running %s" % (
                            ", ".join(resource_usage_str), cmd_str))
                    else:
                        logger.info(
                            "Drmaa command successfully ran %s" % cmd_str)
                except:
                    logger.info("Drmaa command used %s in running %s" %
                                (job_info.resourceUsage, cmd_str))

    #   clean up job template
    drmaa_session.deleteJobTemplate(job_template)

    #   Cleanup job script unless retain_job_scripts is set
    if retain_job_scripts:
        # job scripts have the jobid as an extension
        os.rename(job_script_path, job_script_path + ".%s" % jobid)
    else:
        try:
            os.unlink(job_script_path)
        except OSError:
            if logger:
                logger.warning(
                    "Temporary job script wrapper '%s' missing (and ignored) at clean-up" % job_script_path)

    return stdout, stderr


def enqueue_output(out, queue, echo):
    for line in iter(out.readline, ''):
        queue.put(line)
        if echo is not None:
            echo.write(line)
            echo.flush()
    out.close()


def run_job_locally(cmd_str, logger=None, job_environment=None, working_directory=None, local_echo=False):
    """
    Runs specified command locally instead of drmaa
    """

    popen_params = {"args": cmd_str,
                    "cwd": working_directory if working_directory is not None else os.getcwd(),
                    "shell": True,
                    "stdin": subprocess.PIPE,
                    "stdout": subprocess.PIPE,
                    "stderr": subprocess.PIPE,
                    "bufsize": 1,
                    "universal_newlines": True,
                    "close_fds": ON_POSIX}
    if job_environment is not None:
        popen_params["env"] = job_environment

    process = subprocess.Popen(**popen_params)
    stderrQ = Queue()
    stdoutQ = Queue()
    stdout_t = threading.Thread(target=enqueue_output, args=(
        process.stdout, stdoutQ, sys.stdout if local_echo else None))
    stderr_t = threading.Thread(target=enqueue_output, args=(
        process.stderr, stderrQ, sys.stderr if local_echo else None))
    # if daemon = False, sub process cannot be interrupted by Ctrl-C
    stdout_t.daemon = True
    stderr_t.daemon = True
    stdout_t.start()
    stderr_t.start()
    process.wait()
    stdout_t.join()
    stderr_t.join()
    process.stdin.close()
    process.stdout.close()
    process.stderr.close()

    stdout, stderr = [], []
    try:
        while True:
            stdout.append(stdoutQ.get(False))
    except:
        pass

    try:
        while True:
            stderr.append(stderrQ.get(False))
    except:
        pass

    if process.returncode != 0:
        raise error_drmaa_job("The locally run command was terminated by signal %i:\n"
                              "The original command was:\n%s\n"
                              "The stderr was: \n%s\n\n"
                              "The stdout was: \n%s\n\n" %
                              (-process.returncode, cmd_str, "".join(stderr), "".join(stdout)))

    return stdout, stderr


def touch_output_files(cmd_str, output_files, logger=None):
    """
    Touches output files instead of actually running the command string
    """

    if not output_files or not len(output_files):
        if logger:
            logger.debug("No output files to 'touch' for command:\n%s")
        return

    # make sure is list
    ltypes = (list, tuple)
    if not isinstance(output_files, ltypes):
        output_files = [output_files]
    else:
        output_files = list(output_files)

    #
    #    flatten list of file names
    #    from http://rightfootin.blogspot.co.uk/2006/09/more-on-python-flatten.html
    #
    i = 0
    while i < len(output_files):
        while isinstance(output_files[i], ltypes):
            if not output_files[i]:
                output_files.pop(i)
                i -= 1
                break
            else:
                output_files[i:i + 1] = output_files[i]
        i += 1

    for f in output_files:
        # ignore non strings
        if not isinstance(f, path_str_type):
            continue

        # create file
        if not os.path.exists(f):
            # this should be guaranteed to close the new file immediately?
            with open(f, 'w') as p:
                pass

        # touch existing file
        else:
            os.utime(f, None)


def run_job(cmd_str, job_name=None, job_other_options=None, job_script_directory=None,
            job_environment=None, working_directory=None, logger=None,
            drmaa_session=None, retain_job_scripts=False,
            run_locally=False, output_files=None, touch_only=False,
            verbose=0, local_echo=False,
            resubmit=0, pipeline=None):
    """
    Runs specified command either using drmaa, or locally or only in simulation (touch the output files only)
    """

    pipeline = lookup_pipeline(pipeline)
    
    if touch_only:
        touch_output_files(cmd_str, output_files, logger)
        return "", "",

    if run_locally:
        return run_job_locally(cmd_str, logger, job_environment, working_directory, local_echo)

    return run_job_using_drmaa(cmd_str, job_name, job_other_options,
                               job_script_directory, job_environment, working_directory,
                               retain_job_scripts, logger, drmaa_session,
                               verbose, resubmit, pipeline)



#!/usr/bin/env python
from __future__ import print_function
import unittest
import ruffus
from ruffus import transform, regex, pipeline_run, Pipeline, originate, mkdir
import sys

"""

    test_transform_with_no_re_matches.py

        test messages with no regular expression matches

"""


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]

print("    Ruffus Version = ", ruffus.__version__)


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
@mkdir(tempdir)
@originate(tempdir + "a")
def task_1(o):
    open(o, 'w').close()


@transform(task_1, regex("b"), "task_2.output")
def task_2(i, o):
    for f in o:
        with open(f, 'w') as oo:
            pass


class t_save_to_str_logger:
    """
    Everything to stderr
    """

    def __init__(self):
        self.info_str = ""
        self.warning_str = ""
        self.debug_str = ""

    def info(self, message):
        self.info_str += message

    def warning(self, message):
        self.warning_str += message

    def debug(self, message):
        self.debug_str += message


class Test_task_mkdir(unittest.TestCase):

    def setUp(self):
        """
        """
        pass

    def tearDown(self):
        """
        """
        for d in ['a']:
            fullpath = os.path.join(os.path.dirname(__file__), tempdir + d)
            os.unlink(fullpath)
        os.rmdir(tempdir)

    def test_no_re_match(self):

        save_to_str_logger = t_save_to_str_logger()
        pipeline_run(multiprocess=10, logger=save_to_str_logger,
                     verbose=1, pipeline="main")

        print(save_to_str_logger.warning_str)
        self.assertTrue(
            "no file names matched" in save_to_str_logger.warning_str)
        print("\n    Warning printed out correctly", file=sys.stderr)

    def test_newstyle_no_re_match(self):

        test_pipeline = Pipeline("test")
        test_pipeline.originate(task_1, tempdir + "a").mkdir(tempdir)
        test_pipeline.transform(task_2, task_1, regex("b"), "task_2.output")

        save_to_str_logger = t_save_to_str_logger()
        test_pipeline.run(
            multiprocess=10, logger=save_to_str_logger, verbose=1)
        print(save_to_str_logger.warning_str)
        self.assertTrue(
            "no file names matched" in save_to_str_logger.warning_str)
        print("\n    Warning printed out correctly", file=sys.stderr)


if __name__ == '__main__':
    unittest.main()
